﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace EPiServer.Business.Commerce.Tools.ImportAsset
{
    /// <summary>
    /// Simple implementation of reading a CSV file.
    /// </summary>
    /// <remarks>
    /// Known limitations: does not handle commas in values, even if values are quoted.
    /// </remarks>
    public class CsvReader
    {
        private List<MappingEntry> _entries = new List<MappingEntry>();
        private Func<string[], MappingEntry> _createEntry;

        /// <summary>
        /// Initializes a new instance of the <see cref="CsvReader"/> class.
        /// </summary>
        /// <param name="path">The path to the CSV file.</param>
        public CsvReader(string path)
        {
            using (var stream = File.OpenText(path))
            {
                ReadFirstRow(stream);
                while (!stream.EndOfStream)
                {
                    var entry = ParseRow(stream);
                    if (entry != null)
                    {
                        _entries.Add(entry);
                    }
                }
            }
        }

        /// <summary>
        /// Gets the total entries in the file.
        /// </summary>
        public int TotalEntries
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the product/asset mappings read from the file.
        /// </summary>
        /// <returns>An IEnumerable of MappingEntry containing the information from the CSV file.</returns>
        public IEnumerable<MappingEntry> Entries()
        {
            return _entries;
        }

        /// <summary>
        /// Reads the first row.
        /// </summary>
        /// <param name="stream">The stream.</param>
        /// <remarks>
        /// Figures out which column contains the products and which contains the assets.
        /// </remarks>
        private void ReadFirstRow(StreamReader stream)
        {
            var columns = ReadRow(stream);
            if (columns == null)
            {
                throw new InvalidDataException("The first row must contain at least two headers separated by a comma");
            }

            bool isAssetFirst = columns[0].StartsWith("asset", StringComparison.OrdinalIgnoreCase);
            if (isAssetFirst)
            {
                _createEntry = c => new MappingEntry(c[1], c[0]);
            }
            else
            {
                _createEntry = c => new MappingEntry(c[0], c[1]);
            }
        }

        private MappingEntry ParseRow(StreamReader stream)
        {
            var row = ReadRow(stream);
            if (row == null)
            {
                return null;
            }
            return _createEntry(row);
        }

        private string[] ReadRow(StreamReader stream)
        {
            var row = stream.ReadLine();
            if (String.IsNullOrEmpty(row))
            {
                return null;
            }
            var columns = row.Split(',');
            if (columns.Length < 2)
            {
                return null;
            }
            TotalEntries = TotalEntries + 1;
            return columns;
        }
    }
}
