﻿using System;
using System.Collections.Generic;
using System.Net.Mail;
using System.Threading;
using System.Web;
using System.Web.Configuration;
using EPiServer.Core;
using Mediachase.Commerce.Customers;
using Mediachase.Commerce.Engine.Template;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Dto;
using Mediachase.Commerce.Orders.Managers;
using Mediachase.Commerce.Website;
using Mediachase.Commerce.Website.Helpers;
using Mediachase.Data.Provider;
using EPiServer.Framework.Localization;
using EPiServer.ServiceLocation;
using System.Linq;
using Mediachase.Commerce;

namespace EPiServer.Business.Commerce.Payment.DIBS
{
    public partial class DIBSPayment : System.Web.UI.Page
    {
        private Cart _currentCart = null;
        public const string SessionLatestOrderIdKey = "LatestOrderId";
        public const string SessionLatestOrderNumberKey = "LatestOrderNumber";
        public const string SessionLatestOrderTotalKey = "LatestOrderTotal";
        public const string DIBSSystemName = "DIBS";
        
        private Cart CurrentCart
        {
            get
            {
                if (_currentCart == null)
                {
                    _currentCart = new CartHelper(Cart.DefaultName).Cart;
                }

                return _currentCart;
            }
        }

        private Mediachase.Commerce.Orders.Payment payment;
        private string _acceptUrl;
        private string _cancelUrl;
        private string _callbackUrl;
        private string _orderNumber;

        private Injected<LocalizationService> LocalizationService
        {
            get;
            set;
        }

        /// <summary>
        /// Raises the <see cref="E:System.Web.UI.Control.Load"/> event.
        /// </summary>
        /// <param name="e">The <see cref="T:System.EventArgs"/> object that contains the event data.</param>
        protected override void OnLoad(EventArgs e)
        {
            if (CurrentCart.OrderForms.Count == 0 || CurrentCart.OrderForms[0].Payments.Count == 0)
            {
                return;
            }
            var payments = CurrentCart.OrderForms[0].Payments.ToArray();
            //get DIBS payment by PaymentMethodName, instead of get the first payment in the list, which causes problem when checking out with gift card
            PaymentMethodDto dibs = PaymentManager.GetPaymentMethodBySystemName(DIBSSystemName, SiteContext.Current.LanguageName);
            payment = payments.Where(c => c.PaymentMethodId.Equals(dibs.PaymentMethod.Rows[0]["PaymentMethodId"])).FirstOrDefault();
            if (payment == null)
            {
                return;
            }

            string processingUrl = DIBSPaymentGateway.GetParameterByName(dibs, DIBSPaymentGateway.ProcessingUrl).Value;
            string MD5K1 = DIBSPaymentGateway.GetParameterByName(dibs, DIBSPaymentGateway.MD5Key1).Value;
            string MD5K2 = DIBSPaymentGateway.GetParameterByName(dibs, DIBSPaymentGateway.MD5Key2).Value;

            Providers.StaticUrlProvider urls = new Providers.StaticUrlProvider();

            // Get DIBSPaymentLanding url
            var referencePage = DataFactory.Instance.GetPage(PageReference.StartPage)["DIBSPaymentLandingPage"] as PageReference;
            var landingPage = DataFactory.Instance.GetPage(referencePage);
            _acceptUrl = landingPage != null ? landingPage.LinkURL : string.Empty;

            _cancelUrl = urls.GetUrl(Constants.CheckoutCommandName);
            //In case the user cancels the payment, he'll be redirected back to checkout page.
            _cancelUrl += "#state=step|2&success=false&paymentmethod=dibs";

            _callbackUrl = urls.GetUrl(Constants.DibsPaymentCommandName) ?? string.Empty;

            if (string.IsNullOrEmpty(processingUrl) || string.IsNullOrEmpty(MD5K1) || string.IsNullOrEmpty(MD5K2))
            {
                Cart cart = payment.Parent.Parent as Cart;
                Utilities.UpdateCartInstanceId(cart);
                ProcessUnsuccessfulTransaction(LocalizationService.Service.GetString("/Commerce/shoppingcart/DIBSSettingsError"));
            }

            paymentForm.Action = processingUrl;            

            Response.Cache.SetCacheability(HttpCacheability.NoCache);
            Response.Cache.SetExpires(DateTime.Now.AddSeconds(-1));
            Response.Cache.SetNoStore();
            Response.AppendHeader("Pragma", "no-cache");
            GetValuesFromCookies();
            //Process successful transaction                        
            if (CurrentCart != null && Request.Form["authkey"] != null 
                && Request.Form["amount"] != null && Request.Form["currency"] != null &&
                MD5K1 != null && MD5K2 != null
                && Request.Form["orderid"] != null && Request.Form["transact"] != null)
            {
                string myHash = DIBSPaymentGateway.GetMD5Key(Request.Form["transact"], Request.Form["amount"],
                                                new Currency(Request.Form["currency"]));
                if (myHash.Equals(Request.Form["authkey"]))
                {
                    ProcessSuccessfulTransaction();
                }
            }
            //Process unsuccessful transaction
            else if (CurrentCart != null && Request.Form["merchant"] != null && Request.Form["orderid"] != null
                && Request.Form["amount"] != null && Request.Form["currency"] != null &&
                Request.Form["md5key"] != null)
            {
                string myHash = DIBSPaymentGateway.GetMD5Key(Request.Form["merchant"], Request.Form["orderid"],
                                                new Currency(Request.Form["currency"]), Request.Form["amount"]);
                if (myHash.Equals(Request.Form["md5key"]))
                {
                    ProcessUnsuccessfulTransaction(string.Empty);
                }
            }
            
            if (payment == null)
                Response.Redirect(_cancelUrl);
            else if (payment != null)
            {
                Cart cart = payment.Parent.Parent as Cart;
                // process request from Checkout
                _orderNumber = GenerateOrderNumber(cart);
                string existingValues = CookieHelper.Get("Sample_DIBS_values", true);
                existingValues += Utilities.EncodeCookieValue("orderNumber", _orderNumber);
                CookieHelper.Set("Sample_DIBS_values", existingValues, true);

                Utilities.UpdateCartInstanceId(cart);
            }
            Page.ClientScript.RegisterClientScriptBlock(typeof(DIBSPayment), "submitScript", "window.onload = function (evt) { document.forms[0].submit(); }", true);
            base.OnLoad(e);
        }

        private void GetValuesFromCookies()
        {
            string existingValues = CookieHelper.Get("Sample_DIBS_values", true);
            if (existingValues == null)
            {
                return;
            }
            int index = 0, length = existingValues.Length, delimiterIndex = 0;
            while (index <= length - 1)
            {
                delimiterIndex = existingValues.Substring(index).IndexOf('|');
                string len = existingValues.Substring(index, delimiterIndex);
                index += ++delimiterIndex;
                delimiterIndex = existingValues.Substring(index).IndexOf('|');
                string type = existingValues.Substring(index, delimiterIndex);
                index += ++delimiterIndex;
                string value = existingValues.Substring(index, int.Parse(len));
                index += int.Parse(len);
                if (string.Equals(type, "orderNumber", StringComparison.OrdinalIgnoreCase))
                {
                    _orderNumber = value;
                }
            }
        }

        /// <summary>
        /// Processes the successful transaction.
        /// </summary>
        private void ProcessSuccessfulTransaction()
        {
            PurchaseOrder po = null;
            
            Cart cart = CurrentCart;
            string email = new CartHelper(cart).PrimaryAddress.Email;
            if (cart != null)
            {
                // Make sure to execute within transaction
                using (TransactionScope scope = new TransactionScope())
                {
                    cart.Status = DIBSPaymentGateway.PaymentCompleted;
                    //Rerun workflow to complete 3 activities after process payment activity,
                    //because we break out at the process payment activity when we run this workflow first time
                    OrderGroupWorkflowManager.RunWorkflow(cart, OrderGroupWorkflowManager.CartCheckOutWorkflowName);
                    //Save the transact from DIBS to payment.
                    payment.TransactionID = Request["transact"];
                    payment.AcceptChanges();

                    // Save changes
                    cart.OrderNumberMethod = new Cart.CreateOrderNumber(this.GetOrderNumberFromCookies);
                    //this might cause problem when checkout using multiple shipping address because ECF workflow does not handle it. Modify the workflow instead of modify in this payment
                    po = cart.SaveAsPurchaseOrder();
                    if (CustomerContext.Current.CurrentContact != null)
                    {
                        CustomerContext.Current.CurrentContact.LastOrder = po.Created;
                        CustomerContext.Current.CurrentContact.SaveChanges();
                    }

                    // Remove old cart
                    cart.Delete();
                    cart.AcceptChanges();

                    // Update display name of product by current language
                    po.UpdateDisplayNameWithCurrentLanguage();

                    // Commit changes
                    scope.Complete();
                    CookieHelper.Remove("Sample_DIBS_values");

                }
                _acceptUrl = UriSupport.AddQueryString(_acceptUrl, "success", "true");
                _acceptUrl = UriSupport.AddQueryString(_acceptUrl, "ordernumber", po.TrackingNumber);
                _acceptUrl = UriSupport.AddQueryString(_acceptUrl, "email", email);
                Response.Redirect(_acceptUrl, true);
            }
            else
            {
                Response.Redirect(_cancelUrl, true);
            }
        }

        private void ProcessUnsuccessfulTransaction(string error)
        {
            CookieHelper.Remove("Sample_DIBS_values");
            Response.Redirect(_cancelUrl, true);
        }

        /// <summary>
        /// Gets the oder number from current user's sesion.
        /// </summary>
        /// <param name="cart">The cart.</param>
        /// <returns></returns>
        private string GetOrderNumberFromCookies(Cart cart)
        {
            return _orderNumber;
        }

        /// <summary>
        /// Gets the MD5 authentication key to send to DIBS.
        /// </summary>
        public string MD5Key
        {
            get
            {
                return DIBSPaymentGateway.GetMD5Key(MerchantID, OrderID, Currency, Amount);
            }
        }

        /// <summary>
        /// Generates the order number.
        /// </summary>
        /// <param name="cart">The cart.</param>
        /// <returns></returns>
        private string GenerateOrderNumber(Cart cart)
        {
            string str = new Random().Next(100, 999).ToString();
            return string.Format("PO{0}{1}", cart.OrderGroupId, str);
        }

        /// <summary>
        /// Gets the merchant ID.
        /// </summary>
        /// <value>The merchant ID.</value>
        public string MerchantID
        {
            get
            {
                PaymentMethodDto dibs = PaymentManager.GetPaymentMethodBySystemName("DIBS", SiteContext.Current.LanguageName);
                return DIBSPaymentGateway.GetParameterByName(dibs, DIBSPaymentGateway.UserParameter).Value;
            }
        }

        /// <summary>
        /// Gets the amount.
        /// </summary>
        /// <value>The amount.</value>
        public string Amount
        {
            get
            {
                return (payment != null) ? (payment.Amount*100).ToString("#") : string.Empty;
            }
        }

        /// <summary>
        /// Gets the currency code.
        /// </summary>
        /// <value>The currency code.</value>
        public Currency Currency
        {
            get
            {
                if (payment == null)
                {
                    return string.Empty;
                }
                return string.IsNullOrEmpty(payment.Parent.Parent.BillingCurrency) ?
                    SiteContext.Current.Currency : new Currency(payment.Parent.Parent.BillingCurrency);
            }
        }

        /// <summary>
        /// Gets the order ID.
        /// </summary>
        /// <value>The order ID.</value>
        public string OrderID
        {
            get
            {
                return _orderNumber;
            }
        }

        /// <summary>
        /// Gets the callback URL, which is used by DIBS to call back when
        /// transaction is approved.
        /// </summary>
        /// <value>The callback URL.</value>
        public string CallbackUrl
        {
            get
            {
                return _callbackUrl;
            }
        }

        /// <summary>
        /// Convert the site language to the language which DIBS can support.
        /// </summary>
        /// <value>The language.</value>
        public string Language
        {
            get
            {
                if (SiteContext.Current.LanguageName.StartsWith("da", StringComparison.OrdinalIgnoreCase))
                    return "da";
                else if (SiteContext.Current.LanguageName.StartsWith("sv", StringComparison.OrdinalIgnoreCase))
                    return "sv";
                else if (SiteContext.Current.LanguageName.StartsWith("no", StringComparison.OrdinalIgnoreCase))
                    return "no";
                else if (SiteContext.Current.LanguageName.StartsWith("en", StringComparison.OrdinalIgnoreCase))
                    return "en";
                else if (SiteContext.Current.LanguageName.StartsWith("nl", StringComparison.OrdinalIgnoreCase))
                    return "nl";
                else if (SiteContext.Current.LanguageName.StartsWith("de", StringComparison.OrdinalIgnoreCase))
                    return "de";
                else if (SiteContext.Current.LanguageName.StartsWith("fr", StringComparison.OrdinalIgnoreCase))
                    return "fr";
                else if (SiteContext.Current.LanguageName.StartsWith("fi", StringComparison.OrdinalIgnoreCase))
                    return "fi";
                else if (SiteContext.Current.LanguageName.StartsWith("es", StringComparison.OrdinalIgnoreCase))
                    return "es";
                else if (SiteContext.Current.LanguageName.StartsWith("it", StringComparison.OrdinalIgnoreCase))
                    return "it";
                else if (SiteContext.Current.LanguageName.StartsWith("fo", StringComparison.OrdinalIgnoreCase))
                    return "fo";
                else if (SiteContext.Current.LanguageName.StartsWith("pl", StringComparison.OrdinalIgnoreCase))
                    return "pl";
                return "en";
            }
        }

    }
}
