﻿using System;
using System.Collections;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Web;
using DataCash;
using EPiServer.Framework.Localization;
using EPiServer.ServiceLocation;
using log4net;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Dto;
using Mediachase.Commerce.Orders.Managers;
using Mediachase.Commerce.Plugins.Payment;
using Mediachase.Commerce.Website;
using Mediachase.Commerce.Website.Helpers;

namespace EPiServer.Business.Commerce.Payment.DataCash
{
    public class DataCashPaymentGateway : AbstractPaymentGateway
    {
        #region Public constants

        private static ILog _logger = LogManager.GetLogger(typeof(DataCashPaymentGateway));
        public const string HostAddressParameter = "HostAddressParameter";
        public const string UserIdParameter = "UserIdParameter";
        public const string PasswordParameter = "PasswordParameter";
        public const string PathToLogFileParameter = "PathToLogFileParameter";
        public const string TimeOutParameter = "TimeOutParameter";
        public const string LoggingLevelParameter = "LogginLevelParameter";
        public const string ProxyParameter = "ProxyParameter";
        public const string PaymentPageIdParameter = "PaymentPageId";

        /// <summary>
        /// Represents the session key for storing merchant reference, which is a unique reference number for each transaction (Min 6, max 30 alphanumeric character)
        /// </summary>
        public const string MerchantReferenceSessionKey = "MerchantReference";

        /// <summary>
        /// The unique system name of DataCash payment module, which has been configured Payment section in Commerce Manager
        /// </summary>
        public const string PaymentSystemName = "DataCash";

        /// <summary>
        /// Represents the session key for storing Reference key
        /// </summary>
        public const string ReferenceSessionKey = "DataCashReference";

        #endregion

        #region Properties

        /// <summary>
        /// Gets or sets the data cash configuration.
        /// </summary>
        /// <value>The data cash configuration.</value>
        public Config DataCashConfiguration { get; private set; }

        /// <summary>
        /// Gets or sets the data cash payment method dto.
        /// </summary>
        /// <value>The data cash payment method dto.</value>
        public PaymentMethodDto DataCashPaymentMethodDto { get; private set; }

        /// <summary>
        /// Gets or sets the user id.
        /// </summary>
        /// <value>The user id.</value>
        public string UserId { get; private set; }

        /// <summary>
        /// Gets or sets the password.
        /// </summary>
        /// <value>The password.</value>
        public string Password { get; private set; }

        #endregion

        /// <summary>
        /// Initializes a new instance of the <see cref="DataCashPaymentGateway"/> class.
        /// </summary>
        public DataCashPaymentGateway()
        {
            Init();
        }

        /// <summary>
        /// Loads the configuration data.
        /// </summary>
        private void Init()
        {
            DataCashPaymentMethodDto = PaymentManager.GetPaymentMethodBySystemName(PaymentSystemName, SiteContext.Current.LanguageName);

            UserId = DataCashPaymentMethodDto.GetParameterValueAsString(UserIdParameter);
            Password = DataCashPaymentMethodDto.GetParameterValueAsString(PasswordParameter);
            DataCashConfiguration = GetConfig(DataCashPaymentMethodDto);
        }

        /// <summary>
        /// Processes the payment.
        /// </summary>
        /// <param name="payment">The payment.</param>
        /// <param name="message">The message.</param>
        /// <remarks>When "Complete" or "Refund" shipment in Commerce Manager, this method will be called again with the TransactionType is Capture/Credit</remarks>
        /// <returns></returns>
        public override bool ProcessPayment(Mediachase.Commerce.Orders.Payment payment, ref string message)
        {
            // When a refund is created by return process, 
            // this method will be called again with the TransactionType is Creadit
            if (payment.TransactionType.Equals(TransactionType.Credit.ToString(), StringComparison.OrdinalIgnoreCase))
            {
                // process refund
                return SendRefundRequest(payment, payment.TransactionID, payment.Amount, ref message);
            }
            
            if (payment.TransactionType.Equals(TransactionType.Authorization.ToString(), StringComparison.OrdinalIgnoreCase))
            {
                var session = HttpContext.Current.Session;
                var dataCashReference = session[ReferenceSessionKey] as string;
                var merchantReference = session[MerchantReferenceSessionKey] as string;
                var authenticateCode = string.Empty;
                var result = (PreAuthenticateRequest(payment, dataCashReference, merchantReference, payment.Amount, out authenticateCode, ref message) &&
                        SendFulfillRequest(payment, dataCashReference, authenticateCode, merchantReference, payment.Amount, ref message));
                if (!result && !string.IsNullOrEmpty(message))
                {
                    //just return a generic error
                    _logger.Error(message);
                    message = LocalizationService.GetString("/Commerce/Checkout/DataCash/GenericError");
                }

                payment.TransactionID = dataCashReference;
                return result;
            }

            // return true for other Transaction type
            return true;
        }

        /// <summary>
        /// Creates the pre authenticate request.
        /// </summary>
        /// <param name="payment">The payment.</param>
        /// <param name="datacashReference">The datacash reference.</param>
        /// <param name="merchantReference">The merchant reference.</param>
        /// <param name="amount">The amount.</param>
        /// <param name="authenticateCode">The authenticate code.</param>
        /// <param name="message">The message.</param>
        /// <returns></returns>
        private bool PreAuthenticateRequest(Mediachase.Commerce.Orders.Payment payment, string datacashReference, string merchantReference, decimal amount, out string authenticateCode, ref string message)
        {
            authenticateCode = string.Empty;
            try
            {
                var requestDoc = new Document(DataCashConfiguration);
                requestDoc.set("Request.Authentication.client", UserId);
                requestDoc.set("Request.Authentication.password", Password);
                requestDoc.set("Request.Transaction.TxnDetails.merchantreference", merchantReference);
                requestDoc.set("Request.Transaction.TxnDetails.amount", amount.ToString("0.##"));
                requestDoc.set("Request.Transaction.CardTxn.method", "pre");
                requestDoc.setWithAttributes("Request.Transaction.CardTxn.card_details", datacashReference, new Hashtable() { { "type", "from_hps" } });

                #region 3rd Man Service

                var cartHelper = new CartHelper(payment.Parent.Parent as Cart);
                requestDoc.setWithAttributes("Request.Transaction.TxnDetails.The3rdMan", string.Empty, new Hashtable() { { "type", "realtime" } });

                #region CustomerInformation

                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.CustomerInformation.customer_reference", DateTime.Now.Ticks.ToString());
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.CustomerInformation.forename", cartHelper.PrimaryAddress.FirstName);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.CustomerInformation.surname", cartHelper.PrimaryAddress.LastName);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.CustomerInformation.telephone", cartHelper.PrimaryAddress.DaytimePhoneNumber);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.CustomerInformation.email", cartHelper.PrimaryAddress.Email);
                // for realtime fraud check requests
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.CustomerInformation.mobile_telephone_number", cartHelper.PrimaryAddress.DaytimePhoneNumber);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.CustomerInformation.ip_address", HttpContext.Current.Request.GetIPAddress());

                #endregion

                #region DeliveryAddress

                var shippingAddress = cartHelper.FindAddressByName(payment.Parent.Shipments[0].ShippingAddressId);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.DeliveryAddress.street_address_1", shippingAddress.Line1);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.DeliveryAddress.street_address_2", shippingAddress.Line2);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.DeliveryAddress.city", shippingAddress.City);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.DeliveryAddress.country", GetNumericCountryCode(shippingAddress.CountryCode));
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.DeliveryAddress.postcode", shippingAddress.PostalCode);

                #endregion

                #region BillingAddress

                var billingAddress = cartHelper.FindAddressByName(payment.Parent.BillingAddressId);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.BillingAddress.street_address_1", billingAddress.Line1);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.BillingAddress.street_address_2", billingAddress.Line2);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.BillingAddress.city", billingAddress.City);
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.BillingAddress.country", GetNumericCountryCode(billingAddress.CountryCode));
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.BillingAddress.postcode", billingAddress.PostalCode);

                #endregion

                #region OrderInformation

                requestDoc.setWithAttributes("Request.Transaction.TxnDetails.The3rdMan.OrderInformation.Products", string.Empty, new Hashtable() { { "count", payment.Parent.LineItems.Count.ToString() } });

                for (int i = 0; i < payment.Parent.LineItems.Count; i++)
                {
                    var item = payment.Parent.LineItems[i];
                    var xmlElement = requestDoc.getXmlDocument().CreateElement("Product");

                    var childElement = requestDoc.getXmlDocument().CreateElement("code");
                    childElement.InnerText = item.CatalogEntryId;
                    xmlElement.AppendChild(childElement);

                    childElement = requestDoc.getXmlDocument().CreateElement("quantity");
                    childElement.InnerText = ((int)item.Quantity).ToString();
                    xmlElement.AppendChild(childElement);

                    //recalculate final unit price after all kind of discounts are subtracted from item.ListPrice
                    var finalunitPrice = Math.Round(item.ExtendedPrice / item.Quantity, 2);
                    childElement = requestDoc.getXmlDocument().CreateElement("price");
                    childElement.InnerText = finalunitPrice.ToString("0.##");
                    xmlElement.AppendChild(childElement);

                    requestDoc.setXmlElement("Request.Transaction.TxnDetails.The3rdMan.OrderInformation.Products.Product", xmlElement);
                }

                #endregion

                #region Realtime

                // Comma separated list of original http headers. The headers of most interest are Referrer, Host, Server, User-Agent, Via and X-Forwarded-For, max 200 chars
                var headers = HttpContext.Current.Request.Headers;
                var headerToGet = new string[] { "Referrer", "Host", "Server", "User-Agent", "Via", "X-Forwarded-For" };
                var headersDictionary = headers.Cast<string>().Select(headerKey => new { Key = headerKey, Value = headers[headerKey] }).
                                        Where(header => headerToGet.Any(item => item.Equals(header.Key, StringComparison.OrdinalIgnoreCase)));
                var comma = ",";
                var headersString = string.Join(comma, (from item in headersDictionary select string.Format("{0}:{1}", item.Key, item.Value)).ToArray());
                if (headersString.Length > 200)
                {
                    headersString = headersString.Substring(0, 200).Substring(0, headersString.LastIndexOf(comma));
                }
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.http_header_fields", headersString);
                // Register the consumer associated with this transaction for the consumer product
                requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.register_consumer_watch", "true");

                // Uncomment this line if you are using full account, since 3rd man fraud checking is not available for test account
                //requestDoc.set("Request.Transaction.TxnDetails.The3rdMan.Realtime.real_time_sha1", HashCode(merchantReference));

                #endregion

                #endregion

                var responseDoc = DataCashPaymentGateway.SendTransaction(requestDoc);
                if (responseDoc.IsSucceed())
                {
                    // get recommendation action from 3rd man service:
                    var responseCode = responseDoc.get("Response.The3rdManRealtime.recommendation"); // 0,1,2 - Release,Hole,Reject
                    if (int.Parse(responseCode) == 0)
                    {
                        message = string.Empty;
                        authenticateCode = responseDoc.get("Response.CardTxn.authcode");
                        return !string.IsNullOrEmpty(authenticateCode);
                    }
                    else
                    {
                        message = LocalizationService.GetString("/Commerce/Checkout/DataCash/PaymentRejectedBy3rdMan");
                        return false;
                    }
                }
                else
                {
                    message = responseDoc.GetErrorMessage();
                    return false;
                }
            }
            catch (System.Exception e)
            {
                message = e.Message;
                return false;
            }
        }

        /// <summary>
        /// Generate the SHA-1 hash of a string.
        /// </summary>
        /// <param name="str">The string.</param>
        /// <returns></returns>
        private static string HashCode(string text)
        {
            var data = Encoding.ASCII.GetBytes(text);
            var hashData = new SHA1Managed().ComputeHash(data);
            var hash = string.Empty;
            foreach (var b in hashData)
            {
                hash += b.ToString("X2");
            }
            return hash;
        }

        /// <summary>
        /// Creates the fulfill request.
        /// </summary>
        /// <param name="payment">The payment.</param>
        /// <param name="datacashReference">The datacash reference.</param>
        /// <param name="authenticateCode">The authenticate code.</param>
        /// <param name="merchantReference">The merchant reference.</param>
        /// <param name="amount">The amount.</param>
        /// <param name="message">The message.</param>
        /// <returns></returns>
        private bool SendFulfillRequest(Mediachase.Commerce.Orders.Payment payment, string datacashReference, string authenticateCode, string merchantReference, decimal amount, ref string message)
        {
            try
            {
                var requestDoc = new Document(DataCashConfiguration);
                requestDoc.set("Request.Authentication.client", UserId);
                requestDoc.set("Request.Authentication.password", Password);
                requestDoc.set("Request.Transaction.TxnDetails.merchantreference", merchantReference);
                requestDoc.set("Request.Transaction.TxnDetails.amount", amount.ToString("0.##"));
                requestDoc.set("Request.Transaction.HistoricTxn.method", "fulfill");
                requestDoc.set("Request.Transaction.HistoricTxn.authcode", authenticateCode);
                requestDoc.set("Request.Transaction.HistoricTxn.reference", datacashReference);

                var responseDoc = DataCashPaymentGateway.SendTransaction(requestDoc);
                if (responseDoc.IsSucceed())
                {
                    return true;
                }
                else
                {
                    message = responseDoc.GetErrorMessage();
                    return false;
                }
            }
            catch (System.Exception e)
            {
                message = e.Message;
                return false;
            }
        }

        /// <summary>
        /// Sends the refund request to DataCash service.
        /// </summary>
        /// <param name="payment">The payment.</param>
        /// <param name="datacashReference">The datacash reference.</param>
        /// <param name="amount">The amount of money.</param>
        /// <param name="message">The message.</param>
        /// <returns>True if success</returns>
        private bool SendRefundRequest(Mediachase.Commerce.Orders.Payment payment, string datacashReference, decimal amount, ref string message)
        {
            try
            {
                var requestDoc = new Document(DataCashConfiguration);
                requestDoc.set("Request.Authentication.client", UserId);
                requestDoc.set("Request.Authentication.password", Password);
                requestDoc.set("Request.Transaction.TxnDetails.amount", amount.ToString("0.##"));
                requestDoc.set("Request.Transaction.HistoricTxn.method", "txn_refund");
                requestDoc.set("Request.Transaction.HistoricTxn.reference", datacashReference);

                var responseDoc = DataCashPaymentGateway.SendTransaction(requestDoc);
                if (responseDoc.IsSucceed())
                {
                    return true;
                }
                else
                {
                    message = responseDoc.GetErrorMessage();
                    return false;
                }
            }
            catch (System.Exception e)
            {
                message = e.Message;
                return false;
            }
        }

        #region Static methods

        /// <summary>
        /// Gets the config object from payment configuration.
        /// </summary>
        /// <param name="paymentMethodDto">The payment method dto.</param>
        /// <returns></returns>
        private static Config GetConfig(PaymentMethodDto paymentMethodDto)
        {
            var cfg = new Config();
            cfg.setHost(paymentMethodDto.GetParameterValueAsString(HostAddressParameter));

            var val = string.Empty;
            int intResult = 0;
            val = paymentMethodDto.GetParameterValueAsString(PathToLogFileParameter);
            if (!string.IsNullOrEmpty(val))
            {
                cfg.setLogfile(val);

                // log level can be set only if log file path was set
                intResult = 5;
                int.TryParse(paymentMethodDto.GetParameterValueAsString(LoggingLevelParameter), out intResult);
                cfg.setLogLevel(intResult);
            }

            val = paymentMethodDto.GetParameterValueAsString(ProxyParameter);
            if (!string.IsNullOrEmpty(val))
            {
                cfg.setProxy(val);
            }

            intResult = 60;
            int.TryParse(paymentMethodDto.GetParameterValueAsString(TimeOutParameter), out intResult);
            if (intResult <= 0)
            {
                intResult = 60;
            }
            cfg.setTimeout(intResult);

            return cfg;
        }

        /// <summary>
        /// Sends the transaction to DataCash server.
        /// </summary>
        /// <param name="documentToSend">The document to send.</param>
        /// <returns></returns>
        public static Document SendTransaction(Document documentToSend)
        {
            var gateway = new DataCashPaymentGateway();
            return new Agent(gateway.DataCashConfiguration).send(documentToSend);
        }

        /// <summary>
        /// Gets the numeric contry code.
        /// <see cref="http://unstats.un.org/unsd/methods/m49/m49alpha.htm"/>
        /// </summary>
        /// <param name="countryCode">The country code as ISO ALPHA-3 code.</param>
        /// <returns></returns>
        private static string GetNumericCountryCode(string countryCode)
        {
            if (_countriesCodes == null || _countriesCodes.Count() < 1)
            {
                var filePath = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, @"Templates\Enoteca\Units\Placeable\plugins\payment\DataCash\CountriesCodes.txt");
                _countriesCodes = File.ReadAllLines(filePath);
            }
            return (from line in _countriesCodes
                    let codes = line.Split(';')
                    where countryCode.Equals(codes[1], StringComparison.OrdinalIgnoreCase)
                    select codes[2]).FirstOrDefault();
        }

        /// <summary>
        /// Variable, which represents the content of ContriesCodes file.
        /// </summary>
        private static string[] _countriesCodes;

        #endregion

        private static LocalizationService LocalizationService
        {
            get { return ServiceLocator.Current.GetInstance<LocalizationService>(); }
        }
    }
}
