﻿using System.Linq;
using DataCash;
using Mediachase.Commerce.Orders.Dto;

namespace EPiServer.Business.Commerce.Payment.DataCash
{
    public static class ExtensionMethods
    {
        /// <summary>
        /// Gets the parameter from its name.
        /// </summary>
        /// <param name="paymentMethodDto">The payment method dto.</param>
        /// <param name="parameterName">Name of the parameter.</param>
        /// <returns></returns>
        public static PaymentMethodDto.PaymentMethodParameterRow GetParameterByName(this PaymentMethodDto paymentMethodDto, string parameterName)
        {
            var rowArray = (PaymentMethodDto.PaymentMethodParameterRow[])paymentMethodDto.PaymentMethodParameter.Select(string.Format("Parameter = '{0}'", parameterName));
            if ((rowArray != null) && (rowArray.Length > 0))
            {
                return rowArray[0];
            }

            return null;
        }

        /// <summary>
        /// Gets the parameter value as string, based on its name.
        /// </summary>
        /// <param name="paymentMethodDto">The payment method dto.</param>
        /// <param name="name">The name.</param>
        /// <returns></returns>
        public static string GetParameterValueAsString(this PaymentMethodDto paymentMethodDto, string name)
        {
            var param = GetParameterByName(paymentMethodDto, name);
            return param != null ? param.Value : string.Empty;
        }

        /// <summary>
        /// Get the response status from responseDoc, to determine whether the requested command was succeed or not.
        /// </summary>
        /// <param name="responseDoc">The response doc to check.</param>
        /// <returns>
        /// 	<c>true</c> if the requested command was succeed; otherwise, <c>false</c>.
        /// </returns>
        public static bool IsSucceed(this Document responseDoc)
        {
            if (responseDoc == null)
            {
                return false;
            }
            int returnCode = int.Parse(responseDoc.get("Response.status"));
            return returnCode == 1;
        }

        /// <summary>
        /// Gets the error message from responseDoc.
        /// </summary>
        /// <param name="responseDoc">The response doc.</param>
        /// <returns></returns>
        public static string GetErrorMessage(this Document responseDoc)
        {
            if (responseDoc == null)
            {
                return string.Empty;
            }
            var message = responseDoc.get("Response.information");
            if (string.IsNullOrEmpty(message))
            {
                message = responseDoc.get("Response.reason");
            }
            return message;
        }

        /// <summary>
        /// Gets client IP v4 address.
        /// </summary>
        /// <param name="request">The Http request object.</param>
        /// <returns></returns>
        public static string GetIPAddress(this System.Web.HttpRequest request)
        {
            return GetIPAddress(request.ServerVariables["HTTP_VIA"],
                                request.ServerVariables["HTTP_X_FORWARDED_FOR"],
                                request.ServerVariables["REMOTE_ADDR"]);
        }

        /// <summary>
        /// Gets the user's IP v4 Address
        /// </summary>
        /// <param name="httpVia">HTTP_VIA Server variable</param>
        /// <param name="httpXForwardedFor">HTTP_X_FORWARDED Server variable</param>
        /// <param name="RemoteAddr">REMOTE_ADDR Server variable</param>
        /// <returns>user's IP Address</returns>
        private static string GetIPAddress(string HttpVia, string HttpXForwardedFor, string RemoteAddr)
        {
            string result = string.Empty;
            // Web user - if using proxy
            if (!string.IsNullOrEmpty(HttpVia))
            {
                result = HttpXForwardedFor;
            }
            else // Web user - not using proxy or can't get the Client IP
            {
                result = RemoteAddr;
            }

            // If we can't get a V4 IP from the above, try host address list for internal users.
            if (!IsIPV4(result))
            {
                var ipAddress = (from ip in System.Net.Dns.GetHostAddresses(System.Net.Dns.GetHostName()) where IsIPV4(ip) select ip.ToString()).FirstOrDefault();
                if (!string.IsNullOrEmpty(ipAddress))
                {
                    result = ipAddress;
                }
            }

            return result;
        }

        /// <summary>
        /// Determines whether an IP Address is V4
        /// </summary>
        /// <param name="input">input string</param>
        /// <returns>Is IPV4 True or False</returns>
        private static bool IsIPV4(string input)
        {
            bool result = false;
            System.Net.IPAddress address = null;

            if (System.Net.IPAddress.TryParse(input, out address))
            {
                result = IsIPV4(address);
            }
            return result;
        }

        /// <summary>
        /// Determines weather an IP Address is V4
        /// </summary>
        /// <param name="address">input IP address</param>
        /// <returns>Is IPV4 True or False</returns>
        private static bool IsIPV4(System.Net.IPAddress address)
        {
            return address.AddressFamily == System.Net.Sockets.AddressFamily.InterNetwork;
        }
    }
}
