using System;
using System.Collections;
using System.Linq;
using System.Web.UI.HtmlControls;
using DataCash;
using EPiServer.Core;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Website;
using Mediachase.Commerce.Website.BaseControls;
using Mediachase.Commerce.Website.Helpers;
using EPiServer.Framework.Localization;
using EPiServer.ServiceLocation;
using Mediachase.Commerce;

namespace EPiServer.Business.Commerce.Payment.DataCash
{
    /// <summary>
    ///	Implements User interface for generic gateway
    /// </summary>
    public partial class PaymentMethod : BaseStoreUserControl, IPaymentOption
    {
        private DataCashPaymentGateway _gateway = new DataCashPaymentGateway();

        #region Properties

        /// <summary>
        /// Source of the IFrame to display DataCash page
        /// </summary>
        protected string FrameSource { get; set; }

        /// <summary>
        /// Gets the return URL(after DataCash has been captured user data).
        /// </summary>
        /// <value>The return URL.</value>
        private string ReturnUrl
        {
            get
            {
                var link = new UrlBuilder(UriSupport.AbsoluteUrlBySettings("~/Templates/Enoteca/Units/Placeable/plugins/payment/DataCash/DataCashLandingPage.aspx"));
                var page = Page as PageBase;
                if (page != null && page.CurrentPage != null)
                {
                    link.QueryCollection.Add("epslanguage", page.CurrentPage.LanguageID);
                }
                return link.ToString();
            }
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Registers the CSS file, using key to make sure it won't be register twice.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <param name="rootRelativePath">The root relative path.</param>
        private void RegisterCSSFile(string key, string rootRelativePath)
        {
            var header = Page.Header;
            if (header != null)
            {
                // if the style is not registered under key, we register it once
                if (header.FindControl(key) == null)
                {
                    var link = new HtmlLink() { ID = key };
                    link.Attributes.Add("href", rootRelativePath);
                    link.Attributes.Add("rel", "stylesheet");
                    link.Attributes.Add("type", "text/css");
                    header.Controls.Add(link);
                }
            }
        }

        /// <summary>
        /// Gets the currency code from cart.
        /// </summary>
        /// <param name="payment">The payment.</param>
        /// <returns></returns>
        private Currency GetCurrency(CartHelper cartHelper)
        {
            Mediachase.Commerce.Orders.Payment payment = cartHelper.Cart.OrderForms[0].Payments.Count > 1 ? cartHelper.Cart.OrderForms[0].Payments[0] : null;
            if (payment == null || string.IsNullOrEmpty(payment.Parent.Parent.BillingCurrency))
            {
                return SiteContext.Current.Currency;
            }
            else
            {
                return new Currency(payment.Parent.Parent.BillingCurrency);
            }
        }

        /// <summary>
        /// Validates the captured data.
        /// </summary>
        /// <param name="message">The error message if there was an error.</param>
        /// <returns></returns>
        private bool ValidateCapturedData(out string message)
        {
            try
            {
                var requestDoc = new Document(_gateway.DataCashConfiguration);
                requestDoc.set("Request.Authentication.client", _gateway.UserId);
                requestDoc.set("Request.Authentication.password", _gateway.Password);
                requestDoc.set("Request.Transaction.HistoricTxn.method", "query");
                requestDoc.set("Request.Transaction.HistoricTxn.reference", Session[DataCashPaymentGateway.ReferenceSessionKey] as string);
                var responseDoc = DataCashPaymentGateway.SendTransaction(requestDoc);
                if (responseDoc.IsSucceed())
                {
                    message = string.Empty;
                    return true;
                }
                else
                {
                    message = responseDoc.GetErrorMessage();
                    return false;
                }
            }
            catch (System.Exception e)
            {
                message = e.Message;
                return false;
            }
        }

        /// <summary>
        /// Setups the hosted card service session, to retreive frame source to display to user.
        /// </summary>
        /// <param name="message">The message.</param>
        /// <param name="frameSource">The frame source.</param>
        /// <returns></returns>
        private bool SetupHostedCardService(out string message, out string frameSource)
        {
            frameSource = string.Empty;
            try
            {
                var cartHelper = new CartHelper(Cart.DefaultName);

                var requestDoc = new Document(_gateway.DataCashConfiguration);
                requestDoc.set("Request.Authentication.client", _gateway.UserId);
                requestDoc.set("Request.Authentication.password", _gateway.Password);

                var merchRef = DateTime.Now.Ticks.ToString();
                Session[DataCashPaymentGateway.MerchantReferenceSessionKey] = merchRef; // A unique reference number for each transaction (Min 6, max 30 alphanumeric character)
                requestDoc.set("Request.Transaction.TxnDetails.merchantreference", merchRef);
                var amount = GetCartAmount(cartHelper).ToString("0.##");
                requestDoc.setWithAttributes("Request.Transaction.TxnDetails.amount", amount, new Hashtable() { { "currency", GetCurrency(cartHelper).ToString() } });

                requestDoc.set("Request.Transaction.HpsTxn.method", "setup");
                requestDoc.set("Request.Transaction.HpsTxn.page_set_id", _gateway.DataCashPaymentMethodDto.GetParameterValueAsString(DataCashPaymentGateway.PaymentPageIdParameter));
                // set the return Url to our landing page, to automatic click on "Review & Place order" button, instead of force user to click on it
                requestDoc.set("Request.Transaction.HpsTxn.return_url", ReturnUrl);

                var responseDoc = DataCashPaymentGateway.SendTransaction(requestDoc);
                if (responseDoc.IsSucceed())
                {
                    frameSource = string.Format("{0}?HPS_SessionID={1}", responseDoc.get("Response.HpsTxn.hps_url"), responseDoc.get("Response.HpsTxn.session_id"));
                    message = string.Empty;
                    Session[DataCashPaymentGateway.ReferenceSessionKey] = responseDoc.get("Response.datacash_reference");
                    return true;
                }
                else
                {
                    message = responseDoc.GetErrorMessage();
                    return false;
                }
            }
            catch (System.Exception e)
            {
                message = e.Message;
                return false;
            }
        }

        /// <summary>
        /// Get total amount in current cart.
        /// </summary>
        /// <param name="cart">The current cart.</param>
        /// <returns>The total amount.</returns>
        private static decimal GetCartAmount(CartHelper cart)
        {
            return cart.OrderForm.Total;
        }

        #endregion

        /// <summary>
        /// Handles the Load event of the Page control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void Page_Load(object sender, System.EventArgs e)
        {
            RegisterCSSFile("DataCashStyleSheets", ResolveClientUrl("~/Templates/Enoteca/Units/Placeable/plugins/payment/DataCash/ClientResources/DataCash.css"));

            // register the message for required field
            Page.ClientScript.RegisterClientScriptBlock(Page.GetType(), "DataCashRequiredMessage",
                string.Format("DataCashRequiredMessage = '{0}';", ScriptResourceHelper.EscapeForScript(LocalizationService.GetString("/Commerce/Checkout/DataCash/CreditCardEmpty"))), true);

            // register the DataCash validation script:
            Page.ClientScript.RegisterClientScriptInclude(Page.GetType(), "DataCashValidationScript", ResolveClientUrl("~/Templates/Enoteca/Units/Placeable/plugins/payment/DataCash/ClientResources/jquery.validate.datacash.js"));
        }

        /// <summary>
        /// Raises the <see cref="E:System.Web.UI.Control.PreRender"/> event.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> object that contains the event data.</param>
        protected override void OnPreRender(EventArgs e)
        {
            base.OnPreRender(e);

            var errorMessage = string.Empty;
            var frameSource = string.Empty;
            // setup the HCC service, and get out the frame source to display
            if (SetupHostedCardService(out errorMessage, out frameSource))
            {
                FrameSource = frameSource;
                FrameContainer.Visible = true;
                ErrorContainer.Visible = false;
                return;
            }
            ErrorMessage.Text = errorMessage;
            FrameContainer.Visible = false;
            ErrorContainer.Visible = true;
        }

        #region IPaymentOption Members

        /// <summary>
        /// Validates input data for the control. In case of Credit card pre authentication will be the best way to
        /// validate. The error message if any should be displayed within a control itself.
        /// </summary>
        /// <returns>Returns false if validation is failed.</returns>
        public bool ValidateData()
        {
            ErrorContainer.Visible = false;
            Page.Validate(ID);
            if (!Page.IsValid)
            {
                return false;
            }
            
            // query captured data from DataCash to get out validation result:
            var message = string.Empty;
            if(ValidateCapturedData(out message))
            {
                return true;
            }
            else
            {
                ErrorMessage.Text = message;
                ErrorContainer.Visible = true;
                return false;
            }
        }

        /// <summary>
        /// This method is called before the order is completed. This method should check all the parameters
        /// and validate the credit card or other parameters accepted.
        /// </summary>
        /// <param name="form"></param>
        /// <returns>bool</returns>
        public Mediachase.Commerce.Orders.Payment PreProcess(OrderForm form)
        {
            // Create a new payment
            return new OtherPayment {BillingAddressId = form.BillingAddressId, TransactionType = TransactionType.Authorization.ToString()};
        }

        /// <summary>
        /// This method is called after the order is placed. This method should be used by the gateways that want to
        /// redirect customer to their site.
        /// </summary>
        /// <param name="form">The form.</param>
        /// <returns></returns>
        public bool PostProcess(OrderForm form)
        {
            return true;
        }

        #endregion

        private static LocalizationService LocalizationService
        {
            get { return ServiceLocator.Current.GetInstance<LocalizationService>(); }
        }

    }
}
