using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Threading;
using System.Web;
using EPiServer.ServiceLocation;
using Mediachase.Commerce.Catalog;
using Mediachase.Commerce.Catalog.Objects;
using Mediachase.Commerce.Engine;
using Mediachase.Commerce.Marketing;
using Mediachase.Commerce.Marketing.Dto;
using Mediachase.Commerce.Marketing.Managers;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Managers;
using Mediachase.Commerce.Security;
using Mediachase.MetaDataPlus.Configurator;
using Mediachase.Commerce.Pricing;

namespace Mediachase.Commerce.Website.Helpers
{
    /// <summary>
    /// Cart helper class used to simplify cart operations.
    /// The cart is automatically cached in the current Http Context.
    /// </summary>
    public class CartHelper
    {
        /// <summary>
        /// Default name for the cart.
        /// </summary>
        public static string WishListName = "WishList";
        private string _entryDescriptionMetaFieldName = "Description";

        #region Data Members
        private Cart _Cart;
        #endregion

        #region Constructors
        /// <summary>
        /// Initializes a new instance of the <see cref="CartHelper"/> class.
        /// </summary>
        /// <param name="name">The name.</param>
        public CartHelper(string name)
        {
            Guid userId = SecurityContext.Current.CurrentContactId;
            LoadCart(name, userId);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CartHelper"/> class.
        /// </summary>
        /// <param name="name">The name of cart</param>
        /// <param name="userId">The User Id.</param>
        /// <param name="market">The market.</param>
        public CartHelper(string name, Guid userId, IMarket market)
        {
            LoadCart(name, userId, market.MarketId);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CartHelper"/> class.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="userId">The user id.</param>
        public CartHelper(string name, Guid userId)
        {
            LoadCart(name, userId);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CartHelper"/> class.
        /// </summary>
        /// <param name="cart">The cart.</param>
        public CartHelper(Cart cart)
        {
            if (cart == null)
            {
                throw new ArgumentNullException("cart");
            }
            this._Cart = cart;
        }
        #endregion

        #region Public properties
        /// <summary>
        /// Gets the cart.
        /// </summary>
        /// <value>The cart.</value>
        public virtual Cart Cart
        {
            get
            {
                return this._Cart;
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is empty.
        /// </summary>
        /// <value><c>true</c> if this instance is empty; otherwise, <c>false</c>.</value>
        public virtual bool IsEmpty
        {
            get
            {
                foreach (OrderForm orderForm in this.Cart.OrderForms)
                {
                    if (orderForm.LineItems.Count > 0)
                    {
                        return false;
                    }
                }

                return true;
            }
        }

        /// <summary>
        /// Gets the line items.
        /// </summary>
        /// <value>The line items.</value>
        public virtual IEnumerable<LineItem> LineItems
        {
            get
            {
                foreach (OrderForm orderForm in this.Cart.OrderForms)
                {
                    foreach (LineItem lineItem in orderForm.LineItems)
                    {
                        yield return lineItem;
                    }
                }
            }
        }

        /// <summary>
        /// Gets the primary address.
        /// </summary>
        /// <value>The primary address.</value>
        public virtual OrderAddress PrimaryAddress
        {
            get
            {
                if (String.IsNullOrEmpty(OrderForm.BillingAddressId))
                    return null;

                foreach (OrderAddress address in Cart.OrderAddresses)
                {
                    if (address.Name.Equals(OrderForm.BillingAddressId, StringComparison.OrdinalIgnoreCase))
                        return address;
                }

                return null;
            }
        }

        /// <summary>
        /// Gets the order form.
        /// </summary>
        /// <value>The order form.</value>
        public virtual OrderForm OrderForm
        {
            get
            {
                return GetOrderForm();
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is address required.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this instance is address required; otherwise, <c>false</c>.
        /// </value>
        public virtual bool IsAddressRequired
        {
            get
            {
                foreach (OrderForm orderForm in this.Cart.OrderForms)
                {
                    foreach (Shipment shipment in orderForm.Shipments)
                    {
                        if (String.IsNullOrEmpty(shipment.ShippingAddressId))
                        {
                            return true;
                        }
                    }
                }

                return false;
            }
        }

        public string EntryDescriptionMetaFieldName
        {
            get { return _entryDescriptionMetaFieldName; }
            set { _entryDescriptionMetaFieldName = value; }
        }
        #endregion

        #region Public methods
        /// <summary>
        /// Finds the address by id.
        /// </summary>
        /// <param name="addressId">The address id.</param>
        /// <returns></returns>
        public virtual OrderAddress FindAddressById(string addressId)
        {
            if (String.IsNullOrEmpty(addressId))
                return null;

            foreach (OrderAddress address in Cart.OrderAddresses)
            {
                if (address.OrderGroupAddressId == Int32.Parse(addressId))
                    return address;
            }

            return null;
        }

        /// <summary>
        /// Finds the name of the address by.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns></returns>
        public virtual OrderAddress FindAddressByName(string name)
        {
			if (name == null)
				return null;

            foreach (OrderAddress address in Cart.OrderAddresses)
            {
                if (address.Name.ToUpper() == name.ToUpper())
                    return address;
            }

            return null;
        }

        /// <summary>
        /// Gets the total number of items in the basket.
        /// </summary>
        /// <returns>the total number of items in the basket.</returns>
        public virtual decimal GetTotalItemCount()
        {
            decimal totalItemCount = 0;
            foreach (OrderForm orderForm in this.Cart.OrderForms)
            {
                foreach (LineItem lineItem in orderForm.LineItems)
                {
                    totalItemCount += lineItem.Quantity;
                }
            }
            return totalItemCount;
        }
		
        /// <summary>
        /// Gets the default OrderForm.
        /// </summary>
        /// <returns>the default OrderForm</returns>
        public virtual OrderForm GetOrderForm()
        {
            return this.GetOrderForm(null);
        }

        /// <summary>
        /// Gets the named OrderForm.
        /// </summary>
        /// <param name="orderFormName">The name of the OrderForm object to retrieve.</param>
        /// <returns>The named OrderForm.</returns>
        public virtual OrderForm GetOrderForm(string orderFormName)
        {
            if (String.IsNullOrEmpty(orderFormName))
            {
                orderFormName = Cart.DefaultName;
            }

            OrderForm orderForm = this.Cart.OrderForms[orderFormName];
            if (orderForm == null)
            {
                orderForm = new OrderForm();
                orderForm.Name = orderFormName;
                this.Cart.OrderForms.Add(orderForm);
            }
            return orderForm;
        }

        /// <summary>
        /// Deletes the current basket instance from the database.
        /// </summary>
        public virtual void Delete()
        {
            // Remove any reservations
            // Load existing usage Dto for the current order
            PromotionUsageDto usageDto = PromotionManager.GetPromotionUsageDto(0, Guid.Empty, this.Cart.OrderGroupId);

            // Clear all old items first
            if (usageDto.PromotionUsage.Count > 0)
            {
                foreach (PromotionUsageDto.PromotionUsageRow row in usageDto.PromotionUsage)
                {
                    if (row.OrderGroupId == this.Cart.OrderGroupId)
                    {
                        row.Delete(); 
                    }
                }
            }

            // Save the promotion usage
            PromotionManager.SavePromotionUsage(usageDto);

            string name = this.Cart.Name;

            // Delete the cart
            this.Cart.Delete();
        }

        /// <summary>
		/// Adds the entry. Line item's qty will be increased by 1.
		/// </summary>
		/// <param name="entry">The entry.</param>
		public virtual void AddEntry(Entry entry)
		{
			AddEntry(entry, 1, false);
		}

		/// <summary>
		/// Adds the entry. Line item's qty will be increased by 1.
		/// </summary>
		/// <param name="entry">The entry.</param>
		/// <param name="helpersToRemove"></param>
		public virtual void AddEntry(Entry entry, params CartHelper[] helpersToRemove)
		{
			AddEntry(entry, 1, false, helpersToRemove);
		}

		/// <summary>
		/// Adds the entry.
		/// </summary>
		/// <param name="entry">The entry.</param>
        /// <param name="fixedQuantity">If true, lineitem's qty will be set to <paramref name="fixedQuantity"/> value. Otherwise, <paramref name="fixedQuantity"/> will be added to the current line item's qty value.</param>
		public virtual void AddEntry(Entry entry, bool fixedQuantity)
		{
			AddEntry(entry, 1, fixedQuantity);
		}

		/// <summary>
		/// Adds the entry.
		/// </summary>
		/// <param name="entry">The entry.</param>
        /// <param name="fixedQuantity">If true, lineitem's qty will be set to <paramref name="fixedQuantity"/> value. Otherwise, <paramref name="fixedQuantity"/> will be added to the current line item's qty value.</param>
		/// <param name="helpersToRemove"></param>
		public virtual void AddEntry(Entry entry, bool fixedQuantity, params CartHelper[] helpersToRemove)
		{
			AddEntry(entry, 1, fixedQuantity, helpersToRemove);
		}

        /// <summary>
        /// Adds the entry.
        /// </summary>
        /// <param name="entry">The entry.</param>
        /// <param name="quantity">The quantity.</param>
		/// <param name="fixedQuantity">If true, lineitem's qty will be set to <paramref name="quantity"/> value. Otherwise, <paramref name="quantity"/> will be added to the current line item's qty value.</param>
		/// <param name="helpersToRemove">CartHelper(s) from which the item needs to be removed simultaneously with adding it to the current CartHelper.</param>
		public virtual void AddEntry(Entry entry, decimal quantity, bool fixedQuantity, params CartHelper[] helpersToRemove)
        {
            Guid guid = SecurityContext.Current.CurrentContactId; // this should persist in the cookie
            this.Cart.BillingCurrency = SiteContext.Current.Currency;

            OrderForm orderForm = null;
            if (this.Cart.OrderForms.Count == 0) // create a new one
            {
                orderForm = new OrderForm();
                orderForm.Name = this.Cart.Name;
                this.Cart.OrderForms.Add(orderForm);
            }
            else // use first one
            {
                orderForm = this.Cart.OrderForms[0];
            }

            // Add line items
			LineItem lineItem = CreateLineItem(entry, quantity);

            // check if items already exist
            bool found = false;
            foreach (LineItem item in orderForm.LineItems)
            {
                if (item.CatalogEntryId == lineItem.CatalogEntryId)
                {
					item.Quantity = fixedQuantity ? lineItem.Quantity : item.Quantity + lineItem.Quantity;
                    item.ExtendedPrice = lineItem.ListPrice;
                    found = true;
                    break;
                }
            }

            if (!found)
                orderForm.LineItems.Add(CreateLineItem(entry, quantity));

			// remove entry from other helpers if needed
			if (helpersToRemove != null && helpersToRemove.Length > 0)
			{
				foreach (CartHelper ch in helpersToRemove)
				{
					// if entry is in the helper, remove it
					LineItem li = ch.LineItems.FirstOrDefault<LineItem>(item => item.CatalogEntryId == entry.ID);
					if (li != null)
					{
						li.Delete();

						// If helper is empty, remove it from the database
						if (ch.IsEmpty)
							ch.Delete();

						// save changes
						ch.Cart.AcceptChanges();
					}
				}
			}

            // Save cart
            this.Cart.AcceptChanges();
        }

        /// <summary>
        /// Creates the line item.
        /// </summary>
        /// <param name="entry">The entry.</param>
        /// <param name="quantity">The quantity.</param>
        /// <returns></returns>
        private LineItem CreateLineItem(Entry entry, decimal quantity)
        {
            LineItem lineItem = new LineItem();

            // If entry has a parent, add parents name
            if (entry.ParentEntry != null)
            {
                lineItem.DisplayName = String.Format("{0}: {1}", entry.ParentEntry.Name, entry.Name);
                lineItem.ParentCatalogEntryId = entry.ParentEntry.ID;
            }
            else
            {
                lineItem.DisplayName = entry.Name;
                lineItem.ParentCatalogEntryId = String.Empty;
            }

            Price price = StoreHelper.GetSalePrice(entry, quantity);

            ICurrentMarket currentMarketService = ServiceLocator.Current.GetInstance<ICurrentMarket>();
            IPriceService priceService = ServiceLocator.Current.GetInstance<IPriceService>();
            
            MarketId currentMarketId = currentMarketService.GetCurrentMarket().MarketId;
            IPriceValue listPrice = priceService.GetDefaultPrice(currentMarketId, FrameworkContext.Current.CurrentDateTime, new CatalogKey(entry), price.Money.Currency);

            lineItem.CatalogEntryId = entry.ID;
            lineItem.PlacedPrice = price.Money.Amount;
            lineItem.ExtendedPrice = price.Money.Amount;
            lineItem.MaxQuantity = entry.ItemAttributes.MaxQuantity;
            lineItem.MinQuantity = entry.ItemAttributes.MinQuantity;
            lineItem.Quantity = quantity;
			lineItem.WarehouseCode = entry.ItemAttributes.WarehouseCode;

            if (listPrice != null)
            {
                lineItem.ListPrice = listPrice.UnitPrice.Amount;
            }
            else
            {
                lineItem.ListPrice = lineItem.PlacedPrice;
            }
            
            //Populate LineItem description field from the Entry object, if it exists in entry metaclass
            //Entry description metafield is also configurable so it could map another field to the description field . This fixes issue #718
            if (!string.IsNullOrEmpty(EntryDescriptionMetaFieldName))
            {
                MetaClass mc = MetaClass.Load(CatalogContext.MetaDataContext, entry.MetaClassId);
                foreach (MetaField mf in mc.MetaFields)
                {
                    if (mf.Name.Equals(EntryDescriptionMetaFieldName, StringComparison.InvariantCultureIgnoreCase) &&
                        entry.ItemAttributes[EntryDescriptionMetaFieldName] != null)
                    {
                        //description mapped field found - add to LineItem.Description field
                        lineItem.Description = entry.ItemAttributes[EntryDescriptionMetaFieldName].ToString();
                    }
                }
            }

            return lineItem;
        }

        /// <summary>
        /// Runs the workflow and generates the error message for all the warnings.
        /// </summary>
        /// <param name="name">The name.</param>
        public virtual void RunWorkflow(string name)
        {
			WorkflowResults results = OrderGroupWorkflowManager.RunWorkflow(this.Cart, name);
            object warnings = results.OutputParameters["Warnings"];
            if (warnings != null)
            {
                StringDictionary warningsList = warnings as StringDictionary;
                if (warningsList != null)
                {
                    foreach (string warning in warningsList.Values)
                    {
                        ErrorManager.GenerateError(warning);
                    }
                }
            }       
        }

        /// <summary>
        /// Gets all of the cart promotions regardless of promotion type.
        /// </summary>
        /// <returns></returns>
        public IEnumerable<string> GetCartPromotions()
        {
            return GetCartPromotions(null);
        }

        /// <summary>
        /// Gets the cart promotions based on promotion type or All.
        /// </summary>
        /// <param name="promoKey">The promotion group key.</param>
        /// <returns></returns>
		public IEnumerable<string> GetCartPromotions(PromotionGroup.PromotionGroupKey? promoKey)
		{
            IEnumerable<Discount> discounts = null;
            OrderForm[] allOrderForms = new OrderForm[] { this.OrderForm };
            IEnumerable<Shipment> allShipments;
            IEnumerable<LineItem> allLineItems;

            switch (promoKey)
            {
                case PromotionGroup.PromotionGroupKey.Order:
                    discounts = allOrderForms.SelectMany(x => x.Discounts.Cast<Discount>());
                    break;
                case PromotionGroup.PromotionGroupKey.Shipping:
                    allShipments = allOrderForms.SelectMany(x => x.Shipments.ToArray());
                    discounts = allShipments.SelectMany(x => x.Discounts.Cast<Discount>());
                    break;
                case PromotionGroup.PromotionGroupKey.Entry:
                    allLineItems = allOrderForms.SelectMany(x => x.LineItems.ToArray());
                    discounts = allLineItems.SelectMany(x => x.Discounts.Cast<Discount>());
                    break;
                default:
                    // Get all discounts
                    allShipments = allOrderForms.SelectMany(x => x.Shipments.ToArray());
                    allLineItems = allOrderForms.SelectMany(x => x.LineItems.ToArray());
                    discounts = allOrderForms.SelectMany(x => x.Discounts.Cast<Discount>());
                    discounts = discounts.Concat(allShipments.SelectMany(x => x.Discounts.Cast<Discount>()));
                    discounts = discounts.Concat(allLineItems.SelectMany(x => x.Discounts.Cast<Discount>()));
                    break;
            }

			//Grouping discounts by promotion type
			var groupingDiscounts = discounts.GroupBy(x => x.DiscountId);

			PromotionDto dto = PromotionManager.GetPromotionDto();
			foreach (var groupDiscount in groupingDiscounts)
			{
				var promotionId = groupDiscount.Key;
				PromotionDto.PromotionRow promotion = dto.Promotion.SingleOrDefault(x => x.PromotionId == promotionId);
				if (promotion != null)
				{
					var retVal = GetPromotionDisplayName(promotion, Thread.CurrentThread.CurrentCulture.Name);
					if (string.IsNullOrEmpty(retVal))
					{
						retVal = promotion.Name;
					}
					yield return retVal;
				}
			}
		}

		private static string GetPromotionDisplayName(PromotionDto.PromotionRow row, string languageCode)
		{
			PromotionDto.PromotionLanguageRow[] langRows = row.GetPromotionLanguageRows();
			if (langRows != null && langRows.Length > 0)
			{
				foreach (PromotionDto.PromotionLanguageRow lang in langRows)
				{
					if (lang.LanguageCode.Equals(languageCode, StringComparison.OrdinalIgnoreCase))
					{
						return lang.DisplayName;
					}
				}
			}

			return row.Name;
		}

        /// <summary>
        /// Resets this instance. Will clean up line items, remove payments and delete addresses.
        /// The cart needs to be saved in order for changes to be persisted.
        /// </summary>
        public virtual void Reset()
        {
            // Reset shopping cart
            foreach (LineItem lineItem in this.LineItems)
            {
                lineItem.ShippingAddressId = String.Empty;
                lineItem.ShippingMethodId = Guid.Empty;
                lineItem.ShippingMethodName = null;
            }

            foreach (OrderForm orderForm in Cart.OrderForms)
            {
                foreach (Payment orderPayment in orderForm.Payments)
                    orderPayment.Delete();

                foreach (Shipment shipment in orderForm.Shipments)
                    shipment.Delete();
            }

            foreach (OrderAddress orderAddress in Cart.OrderAddresses)
                orderAddress.Delete();

            Cart.AddressId = String.Empty;
        }
        #endregion

        #region Protected Methods
        /// <summary>
        /// Gets the cache key which is used to store current cart in the http context.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="userId">The user id.</param>
        /// <returns></returns>
        protected virtual string GetCacheKey(string name, Guid userId)
        {
            if (name == null)
            {
                throw new ArgumentNullException("name");
            }

            if (userId == Guid.Empty)
            {
                throw new ArgumentNullException("Invalid user id, can't be empty");
            }

            // Cache cart in the current context
            string cartKey = OrderCache.CreateCacheKey("Cart-http", name, userId.ToString());

            return cartKey;
        }


        /// <summary>
        /// Loads the cart. The cart is loaded from current http context if one is present.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="userId">The user id.</param>
        /// <remarks>Uses the market returned from ICurrentMarket for a market id.</remarks>
        protected virtual void LoadCart(string name, Guid userId)
        {
            ICurrentMarket currentMarketService = ServiceLocator.Current.GetInstance<ICurrentMarket>();
            MarketId currentMarketId = currentMarketService.GetCurrentMarket().MarketId;

            LoadCart(name, userId, currentMarketId);
        }

        
        /// <summary>
        /// Loads the cart. The cart is loaded from current http context if one is present.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="userId">The user id.</param>
        /// <param name="marketId">The market id.</param>
        protected virtual void LoadCart(string name, Guid userId, MarketId marketId)
        {
            // Cache cart in the current context
            string cartKey = GetCacheKey(name, userId);

            Cart cart = null;
            if (HttpContext.Current != null)
                cart = HttpContext.Current.Items[cartKey] as Cart;

            // Make sure to reinitialize cart when it was deleted
            if (cart == null || cart.ObjectState == Mediachase.MetaDataPlus.MetaObjectState.Deleted)
            {
                cart = OrderContext.Current.GetCart(name, userId, marketId);

                if (String.IsNullOrEmpty(cart.CustomerName) || cart.CustomerName.Equals(SecurityContext.Anonymous, StringComparison.OrdinalIgnoreCase))
                    cart.CustomerName = SecurityContext.Current.CurrentUserName;

                if (HttpContext.Current != null)
                    HttpContext.Current.Items[cartKey] = cart;
            }

            this._Cart = cart;
        }
        #endregion
    }
}
