﻿using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Linq;
using Mediachase.BusinessFoundation.Data;
using Mediachase.Commerce.Assets;
using Mediachase.Commerce.Catalog;
using Mediachase.Commerce.Catalog.Dto;
using Mediachase.Commerce.Catalog.Managers;
using Mediachase.BusinessFoundation.Data.Business;

namespace EPiServer.Business.Commerce.Tools.ImportAsset
{
    /// <summary>
    /// Base class for different types of asset import implementations.
    /// </summary>
    public abstract class ImportBase
    {
        // By using the Assets response group, we make sure we bring back the mimimum amount of data needed to do asset importing.
        private static CatalogEntryResponseGroup _assetResponseGroup = new CatalogEntryResponseGroup(CatalogEntryResponseGroup.ResponseGroup.Assets);

        // This variable will hold the identity of the asset root folder.
        private PrimaryKeyId _rootId = PrimaryKeyId.Empty;
        private string _assetFolderName;
        private FolderEntity _assetFolder;
        private string[] _assetGroups;

        /// <summary>
        /// Initializes a new instance of the <see cref="ImportBase"/> class.
        /// </summary>
        /// <param name="assetPath">The local file path where the files to import are present.</param>
        /// <param name="assetFolderName">Name of the asset folder that will be used/created in ECF.</param>
        public ImportBase(string assetPath, string assetFolderName)
        {
            AssetPath = assetPath;
            _assetFolderName = assetFolderName;
        }

        /// <summary>
        /// Does the import.
        /// </summary>
        /// <param name="entries">The entries to import.</param>
        public abstract void DoImport(IEnumerable<MappingEntry> entries);

        /// <summary>
        /// Gets the local file path where assets to import are stored.
        /// </summary>
        public string AssetPath
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets all available asset groups.
        /// </summary>
        /// <remarks>
        /// The asset groups are determined by subfolders on the first level below the AssetPath. I e if you have the AssetPath set as "C:\Images" and below 
        /// the folder you have "Small", "Medum", and "Large" then your asset groups will be "Small", "Medium" and "Large".
        /// </remarks>
        public string[] AssetGroups
        {
            get
            {
                if (_assetGroups == null)
                {
                    _assetGroups = Directory.GetDirectories(AssetPath).Select(s => s.Substring(s.LastIndexOf('\\') + 1)).ToArray();
                }
                return _assetGroups;
            }
        }

        /// <summary>
        /// Gets the ECF asset folder where to store imported assets.
        /// </summary>
        public FolderEntity AssetFolder
        {
            get
            {
                if (_assetFolder == null)
                {
                    _assetFolder = GetOrCreateFolderEntity(RootId, _assetFolderName);
                }
                return _assetFolder;
            }
        }

        /// <summary>
        /// Gets the root id for asset folders.
        /// </summary>
        private PrimaryKeyId RootId
        {
            get
            {
                if (_rootId == PrimaryKeyId.Empty)
                {
                    FolderEntity[] root = FolderEntity.ListRootFolders();
                    if (root == null || root.Length > 1)
                    {
                        return PrimaryKeyId.Empty;
                    }

                    _rootId = root[0].PrimaryKeyId.Value;
                }

                return _rootId;
            }
        }

        public FolderEntity GetOrCreateFolderEntity(FolderEntity parentEntity, string folderName)
        {
            return GetOrCreateFolderEntity(parentEntity.PrimaryKeyId.Value, folderName);
        }

        /// <summary>
        /// Gets existing folder or create folder entity if it does not exist.
        /// </summary>
        /// <param name="parentId">The parent id.</param>
        /// <param name="folderName">Name of the folder.</param>
        /// <returns>The FolderEntity.</returns>
        private FolderEntity GetOrCreateFolderEntity(PrimaryKeyId parentId, string folderName)
        {
            // Create a filter expression to get the minimum amount of information
            var filters = new FilterElement[] 
	        {
		        FilterElement.EqualElement("ParentId", parentId),   // Find children of folder with id=parentId
                FilterElement.EqualElement("Name", folderName)      // ...that has the name=folderName
	        };
            
            // Get the matching folders (folders indicated by the first parameter)
            var folderEntity = EntityManager.List<FolderEntity>("Folder", filters).FirstOrDefault();

            // If not found
            if (folderEntity == null)
            {
                // Create a new folder
                folderEntity = CreateFolderEntity(parentId, folderName);
            }
            return folderEntity;
        }

        private FolderEntity CreateFolderEntity(PrimaryKeyId parentId, string folderName)
        {
            // This is a bit messy, since FolderEntity does not support direct creation of new folders

            // Create a folder entity object with hte correct name.
            FolderEntity folderEntity = new FolderEntity { Name = folderName };

            // Call out to Business Foundation to create a new entity of the given type.
            folderEntity.PrimaryKeyId = BusinessManager.Create(folderEntity);

            // Make sure the folder gets placed in the right location.
            FolderEntity.Move(folderEntity, parentId);

            return folderEntity;
        }

        /// <summary>
        /// Gets the asset or creates it if it does not exist.
        /// </summary>
        /// <param name="parentFolderEntity">The parent folder entity.</param>
        /// <param name="assetName">Name of the asset.</param>
        /// <param name="filePath">The file path where the asset to import is located.</param>
        /// <returns>A FolderElementEntity to associate with products/variations.</returns>
        /// <remarks>
        /// This method will use an existing asset in ECF with the same name if it is found. However there is no check that this existing asset contains the same data
        /// as the asset we're trying to import. If you need to re-import assets with the same name, but new content you need to clear out the asset folder first.
        /// </remarks>
        public FolderElementEntity GetOrCreateElementEntity(FolderEntity parentFolderEntity, string assetName, string filePath)
        {
            // Create a filter expression to get the minimum amount of information
            var filters = new FilterElement[] 
	        {
		        FilterElement.EqualElement("ParentId", parentFolderEntity.PrimaryKeyId),    // Find children of folder with id=parentId
                FilterElement.EqualElement("Name", assetName)                               // ...that has name=assetName
	        };

            // Get the matching folder elements (indicated by the "FolderElement" string).
            var asset = EntityManager.List<FolderElementEntity>("FolderElement", filters).FirstOrDefault();

            // if not found...
            if (asset == null)
            {
                // Create/import the asset
                asset = CreateElementEntityFromAssetFile(parentFolderEntity.PrimaryKeyId.Value, assetName, filePath);
            }

            return asset;
        }

        /// <summary>
        /// Imports and creates the asset from asset file.
        /// </summary>
        /// <param name="parentId">The parent id.</param>
        /// <param name="assetName">Name of the asset.</param>
        /// <param name="filePath">The file path.</param>
        /// <returns></returns>
        private static FolderElementEntity CreateElementEntityFromAssetFile(PrimaryKeyId parentId, string assetName, string filePath)
        {
            using (FileStream stream = File.OpenRead(filePath))
            {
                // FolderElementEntity.Create will create an asset containing the data from the stream and put it into the indicated folder.
                return FolderElementEntity.Create(parentId, assetName, stream, Guid.Empty);
            }
        }

        /// <summary>
        /// Associates the assets with aproduct.
        /// </summary>
        /// <param name="assets">The imported assets to associate with  product.</param>
        /// <param name="productDto">The product that should receive the assets.</param>
        public void AssociateAssetsWithProduct(IEnumerable<AssetWrapper> assets, CatalogEntryDto productDto)
        {
            bool isSaveRequired = true;
            foreach (var assetElement in assets)
            {
                // Make sure we do not try to add already existing associations.
                if (IsAlreadyAssociated(productDto, assetElement.AssetKey))
                {
                    continue;
                }
                AddAssociation(productDto, assetElement);
                isSaveRequired = true;
            }

            if (isSaveRequired)
            {
                // Make sure we write back the changes to the database
                CatalogContext.Current.SaveCatalogEntry(productDto);
            }
        }

        private bool IsAlreadyAssociated(CatalogEntryDto productDto, string key)
        {
            // Since we expect the number of associated assets to be rather small (in the range of 2 - 10) it is ok to do this. If we will
            // have a much larger amount of associated assets for each product we should probably move to a more efficient way that does not
            // have to go thru all existing associations.
            return productDto.CatalogItemAsset.Any(r => r.AssetKey.Equals(key, StringComparison.OrdinalIgnoreCase));
        }

        private void AddAssociation(CatalogEntryDto productDto, AssetWrapper assetElement)
        {
            CatalogEntryDto.CatalogItemAssetRow row = productDto.CatalogItemAsset.NewCatalogItemAssetRow();
            row.AssetKey = assetElement.AssetKey;
            row.AssetType = "file";
            row.CatalogEntryId = productDto.CatalogEntry[0].CatalogEntryId;
            row.GroupName = assetElement.Group;
            row.SortOrder = 0;

            if (row.RowState == DataRowState.Detached)
            {
                productDto.CatalogItemAsset.Rows.Add(row);
            }
        }

        /// <summary>
        /// Gets the DTO object for a product/variation from ECF.
        /// </summary>
        /// <param name="productCode">The product code.</param>
        /// <returns>A catalogEntryDto or null if not found.</returns>
        public CatalogEntryDto GetProductDto(string productCode)
        {
            var productDto = CatalogContext.Current.GetCatalogEntryDto(productCode, _assetResponseGroup);
            if (productDto == null || productDto.CatalogEntry.Count <= 0)
            {
                return null;
            }
            return productDto;
        }

        /// <summary>
        /// Gets all valid asset groups for the asset.
        /// </summary>
        /// <param name="assetName">Name of the asset.</param>
        /// <returns>The valid asset groups.</returns>
        /// <remarks>
        /// This handles the case when the mapping file (csv file) contains an entry that maps an asset to a product, but the asset file does not exist in all the different
        /// asset group folders. In that case we will only import the asset that actually exists.
        /// </remarks>
        public IEnumerable<string> ValidAssetGroups(string assetName)
        {
            return AssetGroups.Where(g => File.Exists(FullAssetPath(g, assetName)));
        }

        /// <summary>
        /// Gets the full path to an asset file.
        /// </summary>
        /// <param name="segments">The segments for the asset file.</param>
        /// <returns>A correctly joined file path pointing to the asset file.</returns>
        public string FullAssetPath(params string[] segments)
        {
            var path = AssetPath;
            foreach (var segment in segments)
            {
                path = Path.Combine(path, segment);
            }
            return path;
        }
    }


}
