﻿using System;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Web;
using EPiServer.Core;
using Klarna.Core;
using Mediachase.Commerce.Catalog;
using Mediachase.Commerce.Catalog.Managers;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Dto;
using Mediachase.Commerce.Orders.Managers;
using Mediachase.Commerce.Website;
using Mediachase.Commerce.Website.Helpers;
using Address = EPiServer.Business.Commerce.Payment.Klarna.Address;
using Mediachase.Commerce;

namespace EPiServer.Business.Commerce.Payment.Klarna
{
    /// <summary>
    /// This class intent to support both Klarna Invoice payment and Klarna Part Payment gateway.
    /// </summary>
    internal class KlarnaHelper
    {
        internal const string Swe = "SWE";
        internal const string Fin = "FIN";
        internal const string Nor = "NOR";
        internal const string Dnk = "DNK";
        internal const string Deu = "DEU";
        internal const string Nld = "NLD";
        internal const string Aut = "AUT";

        internal const string SwedishPclasssValues = "SwedishPclasssValues";
        internal const string FinnishPclasssValues = "FinnishPclasssValues";
        internal const string GermanPclasssValues = "GermanPclasssValues";
        internal const string DutchPclasssValues = "DutchPclasssValues";
        internal const string DanishPclasssValues = "DanishPclasssValues";
        internal const string NorwegianPclasssValues = "NorwegianPclasssValues";
        internal const string AustriaPclasssValues = "AustriaPclasssValues";

        internal const string TestFlag = "TestFlag";
        internal const string MerchantId = "MerchantId";
        internal const string SharedSecret = "SharedSecret";
        internal const string EnableCustomerEnterAddress = "EnableCustomerEnterAddress";
        internal const string NeedUpdatePclass = "NeedUpdatePclass";
        internal const string KlarnaPartPayment = "KlarnaPartPayment";
        internal const string KlarnaInvoice = "KlarnaInvoice";

        internal const string PcStorage = "xml";
        internal static readonly string PcUriPartPayment = HttpContext.Current.Server.MapPath("~/Templates/Enoteca/Units/Placeable/Plugins/Payment/KlarnaPartPayment/ClientResources/storage.xml");
        internal static readonly string PcUriInvoice = HttpContext.Current.Server.MapPath("~/Templates/Enoteca/Units/Placeable/Plugins/Payment/KlarnaInvoice/ClientResources/storage.xml");
		
        internal const string SocialSecurityNo = "SocialSecurityNo";
        internal const string GermanyHouseNumberText = "GermanyHouseNumberText";
        internal const string NetherlandsHouseNumberText = "NetherlandsHouseNumberText";
        internal const string NetherlandsHouseExtensionText = "NetherlandsHouseExtensionText";

        internal enum KlarnaPayment
        {
            InvoicePayment,
            PartPayment
        }

        /// <summary>
        /// Gets the klarna country enum by country code
        /// </summary>
        /// <param name="countryCode">The country code.</param>
        /// <returns>The klarna country enum value</returns>
        internal static API.Country GetKlarnaCountry(string countryCode)
        {
            switch (countryCode)
            {
                case Swe:
                    return API.Country.Sweden;
                case Fin:
                    return API.Country.Finland;
                case Nor:
                    return API.Country.Norway;
                case Dnk:
                    return API.Country.Denmark;
                case Deu:
                    return API.Country.Germany;
                case Nld:
                    return API.Country.Netherlands;
                case Aut:
                    return API.Country.Austria;
                default:
                    return API.Country.Sweden;
            }
        }

        /// <summary>
        /// Gets the Klarna country encoding by country code
        /// </summary>
        /// <param name="countryCode">The country code.</param>
        /// <returns>The klarna encoding enum</returns>
        internal static API.Encoding GetKlarnaEncodingByCountry(string countryCode)
        {
            switch (countryCode)
            {
                case Swe:
                    return API.Encoding.Swedish;
                case Fin:
                    return API.Encoding.Finnish;
                case Nor:
                    return API.Encoding.Norwegian;
                case Dnk:
                    return API.Encoding.Danish;
                case Deu:
                    return API.Encoding.German;
                case Nld:
                    return API.Encoding.Dutch;
                case Aut:
                    return API.Encoding.Austria;
                default:
                    return API.Encoding.Swedish;
            }
        }

        /// <summary>
        /// Gets the currency code by country.
        /// </summary>
        /// <param name="countryCode">The country code.</param>
        /// <returns>The currency code string</returns>
        internal static Currency GetCurrencyByCountry(string countryCode)
        {
            switch (countryCode)
            {
                case Swe:
                    return Currency.SEK;
                case Fin:
                    return Currency.EUR;
                case Nor:
                    return Currency.NOK;
                case Dnk:
                    return Currency.DKK;
                case Deu:
                    return Currency.EUR;
                case Nld:
                    return Currency.EUR;
                case Aut:
                    return Currency.EUR;
                default:
                    return Currency.SEK;
            }
        }

        /// <summary>
        /// Gets the store name by country.
        /// </summary>
        /// <param name="countryCode">The country code.</param>
        /// <returns>A store name based on country</returns>
        internal static string GetStoreNameByCountry(string countryCode)
        {
            switch (countryCode)
            {
                case Swe:
                    return SwedishPclasssValues;
                case Fin:
                    return FinnishPclasssValues;
                case Nor:
                    return NorwegianPclasssValues;
                case Dnk:
                    return DanishPclasssValues;
                case Deu:
                    return GermanPclasssValues;
                case Nld:
                    return DutchPclasssValues;
                case Aut:
                    return AustriaPclasssValues;
                default:
                    return string.Empty;
            }
        }

        /// <summary>
        /// Gets the klarna base config. EID, SharedSecret, servermode
        /// </summary>
        /// <param name="klarnaPayment">The klarna payment type </param>
        /// <returns>A klarnaConfig object which store configured parameter</returns>
        internal static KlarnaConfig GetKlarnaBaseConfig(KlarnaPayment klarnaPayment)
        {
            KlarnaConfig klarnaConfig = new KlarnaConfig();

            // MerchantId parameter
            var merchantIdparameter = GetParameterByName(MerchantId, klarnaPayment);
            if (merchantIdparameter != null)
            {
                klarnaConfig.EID = int.Parse(merchantIdparameter.Value);
            }

            // SharedSecret parameter
            var sharedSecretparameter = GetParameterByName(SharedSecret, klarnaPayment);
            if (sharedSecretparameter != null)
            {
                klarnaConfig.Secret = sharedSecretparameter.Value;
            }

            var serverMode = GetParameterByName(TestFlag, klarnaPayment);
            if (serverMode != null)
            {
                bool isTestMode = false;
                bool.TryParse(serverMode.Value, out isTestMode);
                klarnaConfig.Mode = isTestMode ? API.KlarnaServer.Beta : API.KlarnaServer.Live;
            }

            return klarnaConfig;
        }

        /// <summary>
        /// Determines whether [is allow different address].
        /// </summary>
        /// <param name="klarnaPayment">The klarna payment type </param>
        /// <returns>
        /// 	<c>true</c> if [is allow different address]; otherwise, <c>false</c>.
        /// </returns>
        internal static bool IsAllowDifferentAddress(KlarnaPayment klarnaPayment)
        {
            bool result = false;

            // Allow different address parameter
            var allowDifferentAddressParameter = GetParameterByName(EnableCustomerEnterAddress, klarnaPayment);
            if (allowDifferentAddressParameter != null)
            {
                bool.TryParse(allowDifferentAddressParameter.Value, out result);
            }

            return result;
        }


        /// <summary>
        /// Determines whether [is update pclass].
        /// </summary>
        /// <param name="klarnaPayment">The klarna payment type </param>
        /// <returns>
        /// 	<c>true</c> if [is update pclass]; otherwise, <c>false</c>.
        /// </returns>
        internal static bool IsUpdatePclass(KlarnaPayment klarnaPayment)
        {
            bool isUpdate = false;

            // NeedUpdatePclassParameter parameter
            var needUpdatePclass = GetParameterByName(NeedUpdatePclass, klarnaPayment);
            if (needUpdatePclass != null)
            {
                isUpdate = needUpdatePclass.Value.ToUpper().Equals("YES");
                if (isUpdate)
                {
                    // Set parameter to "no"
                    needUpdatePclass.Value = "no";

                    var paymentMethodDto = PaymentManager.GetPaymentMethodBySystemName(KlarnaPartPayment, SiteContext.Current.LanguageName);
                    if (paymentMethodDto == null)
                    {
                        string method = MethodInfo.GetCurrentMethod().Name;
                        ErrorManager.GenerateError(string.Format("Can't get payment method Dto. Error in {0}", method));
                    }
                    // Save the setting
                    PaymentManager.SavePayment(paymentMethodDto);
                }
            }

            return isUpdate;
        }

        /// <summary>
        /// Updates the klarna config by country. Including Country code, language code, currency code, encoding
        /// </summary>
        /// <param name="countryCode">The country code.</param>
        /// <param name="klarnaConfig">The klarna config.</param>
        internal static void UpdateKlarnaConfigByCountry(string countryCode, KlarnaConfig klarnaConfig)
        {
            switch (countryCode)
            {
                case Swe:
                    klarnaConfig.Country = API.Country.Sweden;
                    klarnaConfig.Language = API.Language.Swedish;
                    klarnaConfig.Currency = API.Currency.SEK;
                    klarnaConfig.Encoding = API.Encoding.Swedish;
                    break;
                case Fin:
                    klarnaConfig.Country = API.Country.Finland;
                    klarnaConfig.Language = API.Language.Finnish;
                    klarnaConfig.Currency = API.Currency.EUR;
                    klarnaConfig.Encoding = API.Encoding.Finnish;
                    break;
                case Nor:
                    klarnaConfig.Country = API.Country.Norway;
                    klarnaConfig.Language = API.Language.Norwegian;
                    klarnaConfig.Currency = API.Currency.NOK;
                    klarnaConfig.Encoding = API.Encoding.Norwegian;
                    break;
                case Dnk:
                    klarnaConfig.Country = API.Country.Denmark;
                    klarnaConfig.Language = API.Language.Danish;
                    klarnaConfig.Currency = API.Currency.DKK;
                    klarnaConfig.Encoding = API.Encoding.Danish;
                    break;
                case Deu:
                    klarnaConfig.Country = API.Country.Germany;
                    klarnaConfig.Language = API.Language.German;
                    klarnaConfig.Currency = API.Currency.EUR;
                    klarnaConfig.Encoding = API.Encoding.German;
                    break;
                case Nld:
                    klarnaConfig.Country = API.Country.Netherlands;
                    klarnaConfig.Language = API.Language.Dutch;
                    klarnaConfig.Currency = API.Currency.EUR;
                    klarnaConfig.Encoding = API.Encoding.Dutch;
                    break;
                case Aut:
                    klarnaConfig.Country = API.Country.Austria;
                    klarnaConfig.Language = API.Language.German;
                    klarnaConfig.Currency = API.Currency.EUR;
                    klarnaConfig.Encoding = API.Encoding.Austria;
                    break;
                default:
                    break;
            }
        }

        /// <summary>
        /// Gets the country code from billing address.
        /// </summary>
        /// <returns>A Country code</returns>
        internal static string GetCountryCode()
        {
            var selectedShippingAddress = new CartHelper(Cart.DefaultName).FindAddressByName("Billing Address");

            return (selectedShippingAddress == null) ? Swe : selectedShippingAddress.CountryCode;
        }

        /// <summary>
        /// Configs the Klarna object
        /// </summary>
        /// <param name="klarna">The klarna.</param>
        /// <param name="klarnaPayment">The klarna payment type </param>
        /// <returns>A Klarna object which has been configured</returns>
        internal static API ConfigKlarna(API klarna, KlarnaPayment klarnaPayment)
        {
            KlarnaConfig klarnaConfig = GetKlarnaBaseConfig(klarnaPayment);
            UpdateKlarnaConfigByCountry(GetCountryCode(), klarnaConfig);

            // Configure storage type and storage URI
            klarnaConfig.PCStorage = PcStorage;
            if (klarnaPayment == KlarnaPayment.PartPayment)
            {
                klarnaConfig.PCURI = PcUriPartPayment;
            }
            else
            {
                klarnaConfig.PCURI = PcUriInvoice;
            }

            // Update klarna config
            klarna.Config(klarnaConfig);

            //It's important that we get the end customers IP attached to the Klarna Object
            klarna.ClientIP = HttpContext.Current.Request.UserHostAddress;

            return klarna;
        }

        /// <summary>
        /// Gets the name of the parameter by.
        /// </summary>
        /// <param name="parameterName">Name of the parameter.</param>
        /// <param name="klarnaPayment">The klarna payment type </param>
        /// <returns>The parameter</returns>
        private static PaymentMethodDto.PaymentMethodParameterRow GetParameterByName(string parameterName, KlarnaPayment klarnaPayment)
        {
            string systemName = klarnaPayment == KlarnaPayment.InvoicePayment ? KlarnaInvoice : KlarnaPartPayment;
            var paymentMethodDto = PaymentManager.GetPaymentMethodBySystemName(systemName, SiteContext.Current.LanguageName);
            if (paymentMethodDto == null)
            {
                string method = MethodInfo.GetCurrentMethod().Name;
                ErrorManager.GenerateError(string.Format("Can't get payment method Dto. Error in {0}", method));
            }

            var rowArray = (PaymentMethodDto.PaymentMethodParameterRow[])paymentMethodDto.PaymentMethodParameter.Select(string.Format("Parameter = '{0}'", parameterName));
            if ((rowArray != null) && (rowArray.Length > 0))
            {
                return rowArray[0];
            }

            return null;
        }

        /// <summary>
        /// Sets the klarna address.
        /// </summary>
        /// <param name="klarnaClient">The klarna client.</param>
        /// <param name="billingAddress">The billing address.</param>
        /// <param name="shippingAddress">The shipping address.</param>
        /// <param name="houseNumber">The house number.</param>
        /// <param name="houseExtension">The house extension.</param>
        internal static void SetKlarnaAddress(API klarnaClient, OrderAddress billingAddress, OrderAddress shippingAddress, string houseNumber, string houseExtension)
        {
            // Make a Klarna address
            klarnaClient.SetAddress(new BillingAddress(
                billingAddress.Email,
                billingAddress.EveningPhoneNumber,
                billingAddress.DaytimePhoneNumber,
                billingAddress.FirstName,
                billingAddress.LastName,
                string.Empty,
                billingAddress.Line1,
                billingAddress.PostalCode,
                billingAddress.City,
                GetKlarnaCountry(billingAddress.CountryCode),
                houseNumber,
                houseExtension,
                string.Empty
                ));

            klarnaClient.SetAddress(new ShippingAddress(
                shippingAddress.Email,
                shippingAddress.EveningPhoneNumber,
                shippingAddress.DaytimePhoneNumber,
                shippingAddress.FirstName,
                shippingAddress.LastName,
                string.Empty,
                shippingAddress.Line1,
                shippingAddress.PostalCode,
                shippingAddress.City,
                GetKlarnaCountry(shippingAddress.CountryCode),
                houseNumber,
                houseExtension,
                string.Empty
                ));
        }

        /// <summary>
        /// Gets the item tax in percent.
        /// </summary>
        /// <param name="item">The item.</param>
        /// <param name="addressByName">Name of the address by.</param>
        /// <returns>Return the tax in percent</returns>
        internal static decimal GetItemTax(LineItem item, OrderAddress addressByName)
        {
            var itemTax = 0M;
            // Get the catalogEntryDto
            var catalogEntryDto = CatalogContext.Current.GetCatalogEntryDto(item.CatalogEntryId, new CatalogEntryResponseGroup(CatalogEntryResponseGroup.ResponseGroup.CatalogEntryFull));
            if (catalogEntryDto.CatalogEntry.Count > 0)
            {
                // Get the variationrows
                var variationRows = catalogEntryDto.CatalogEntry[0].GetVariationRows();
                if (variationRows.Length > 0)
                {
                    // Get TaxCategory
                    var taxCategoryNameById = CatalogTaxManager.GetTaxCategoryNameById(variationRows[0].TaxCategoryId);
                    var valueArray = OrderContext.Current.GetTaxes(Guid.Empty, taxCategoryNameById, Thread.CurrentThread.CurrentCulture.Name, addressByName.CountryCode, addressByName.State, addressByName.PostalCode, addressByName.RegionCode, string.Empty, addressByName.City);
                    if (valueArray.Length > 0)
                    {
                        itemTax = (decimal)valueArray.Where(v => v.TaxType == TaxType.SalesTax).Sum(v => v.Percentage);
                    }
                }
            }
            return itemTax;
        }

        /// <summary>
        /// Determines whether [is currency correct] [the specified currency].
        /// </summary>
        /// <param name="currency">The currency.</param>
        /// <returns>
        /// 	<c>true</c> if [is currency correct] [the specified currency]; otherwise, <c>false</c>.
        /// </returns>
        internal static bool IsCurrencyCorrect(Currency currency)
        {
            return SiteContext.Current.Currency.Equals(currency);
        }

        /// <summary>
        /// Gets the currency name by code.
        /// </summary>
        /// <param name="currencyCode">The currency.</param>
        /// <returns>A string currency name</returns>
        internal static Currency GetCurrencyNameByCode(Currency currency)
        {
            switch (currency.CurrencyCode)
            {
                case "SEK":
                    return "Swedish krona";
                case "EUR":
                    return "Euro";
                case "NOK":
                    return "Norwegian krone";
                case "DKK":
                    return "Danish krone";
                default:
                    return currency.CurrencyCode;
            }
        }

        /// <summary>
        /// Determines whether the current call stack is start from specific source type to destination source type.
        /// </summary>
        /// <param name="destinaion">The destinaion type.</param>
        /// <param name="source">The source type.</param>
        /// <returns>
        /// 	<c>true</c> if the current call stack is start from specific source type to destination source type; otherwise, <c>false</c>.
        /// </returns>
        internal static bool IsCallingFrom(Type destinaion, Type source)
        {
            var stackTrace = new StackTrace();
            StackFrame[] stackFrames = stackTrace.GetFrames();

            if (stackFrames == null)
            {
                return false;
            }

            var stackFrameAtDestination = stackFrames.FirstOrDefault(sf => sf.GetMethod().DeclaringType.UnderlyingSystemType == destinaion);
            var stackFrameAtSource = stackFrames.LastOrDefault(sf => sf.GetMethod().DeclaringType.UnderlyingSystemType == source);

            if (stackFrameAtDestination == null || stackFrameAtSource == null)
            {
                return false;
            }

            return stackFrames.ToList().IndexOf(stackFrameAtDestination) < stackFrames.ToList().IndexOf(stackFrameAtSource);
        }

        /// <summary>
        /// Updates the invoice number in Klarna sytem based on given invoice number to new invoice number which is get from purchase order.
        /// </summary>
        /// <param name="klarnaClient">The proxy to communicate with Klarna server.</param>
        /// <param name="processedForm">The processed form, which contains the tracking number to update</param>
        /// <param name="oldInvoiceNumber">The old invoice number.</param>
        /// <returns>True if success</returns>
        internal static bool UpdateKlarnaInvoiceNumber(API klarnaClient, OrderForm processedForm, string oldInvoiceNumber)
        {
            if (processedForm == null || processedForm.Parent == null)
            {
                return false;
            }
            var purchaseOrder = processedForm.Parent as PurchaseOrder;
            if (purchaseOrder == null)
            {
                return false;
            }

            var newInvoiceNumber = purchaseOrder.TrackingNumber;
            if (!String.IsNullOrEmpty(oldInvoiceNumber) && !String.IsNullOrEmpty(newInvoiceNumber))
            {
                try
                {
                    // Update the orderNumber for the invoice
                    klarnaClient.UpdateOrderNumber(oldInvoiceNumber, newInvoiceNumber);
                }
                catch (KlarnaException kex)
                {
                    ErrorManager.GenerateError(kex.FaultString);
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Validates the currency.
        /// </summary>
        /// <param name="countryCode">The country code to validate.</param>
        /// <returns>True if the currency is supported in the country of billing address.</returns>
        internal static bool ValidateCurrency(string countryCode)
        {
            switch (countryCode)
            {
                case Swe:
                case Fin:
                case Nor:
                case Dnk:
                case Deu:
                case Nld:
                case Aut:
                    // Klarna only support for the countries above.
                    if (!IsCurrencyCorrect(GetCurrencyByCountry(countryCode)))
                    {
                        return false;
                    }
                    break;
            }

            return true;
        }

        /// <summary>
        /// Updates the order address.
        /// </summary>
        /// <param name="form">The form contains order address to update.</param>
        /// <param name="klarnaAddress">The klarna address.</param>
        internal static void UpdateOrderAddress(OrderForm form, Address klarnaAddress)
        {
            if (form == null)
            {
                throw new ArgumentNullException("form");
            }

            if (klarnaAddress == null)
            {
                throw new ArgumentNullException("klarnaAddress");
            }

            foreach (var address in form.Parent.OrderAddresses.Cast<OrderAddress>().Where(address => address.Name == form.BillingAddressId))
            {
                address.LastName = klarnaAddress.LastName;
                address.Line1 = klarnaAddress.Street;
                address.City = klarnaAddress.City;
                address.PostalCode = klarnaAddress.Zip;
                form.Parent.AcceptChanges();
                break;
            }
        }
    }
}
