﻿using System;
using System.Linq;
using System.Web;
using Klarna.Core;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Exceptions;
using Mediachase.Commerce.Plugins.Payment;
using Mediachase.Commerce.Website;
using EPiServer.Framework.Localization;
using EPiServer.ServiceLocation;
using log4net;

namespace EPiServer.Business.Commerce.Payment.Klarna.KlarnaInvoice
{
    /// <summary>
    /// KlarnaPaymentGateway class
    /// </summary>
    public class KlarnaPaymentGateway : AbstractPaymentGateway
    {
        private readonly ILog _logger = log4net.LogManager.GetLogger(typeof(KlarnaPaymentGateway));
        private const KlarnaHelper.KlarnaPayment PaymentType = KlarnaHelper.KlarnaPayment.InvoicePayment;

        private LocalizationService LocalizationService
        {
            get { return ServiceLocator.Current.GetInstance<LocalizationService>(); }
        }

        /// <summary>
        /// Processes the payment.
        /// </summary>
        /// <param name="payment">The payment.</param>
        /// <param name="message">The message.</param>
        /// <returns></returns>
        public override bool ProcessPayment(Mediachase.Commerce.Orders.Payment payment, ref string message)
        {
            // active invoice when order is complete
            // when user click complete order in commerce manager the transaction type will be Capture
            if (payment.TransactionType.Equals(TransactionType.Capture.ToString(), StringComparison.OrdinalIgnoreCase))
            {
                try
                {
                    // create a Klarna client
                    var klarnaProxy = KlarnaHelper.ConfigKlarna(new API(), PaymentType);
                    var klarnaTransaction = payment.TransactionID;
                    klarnaProxy.ActivateInvoice(klarnaTransaction);
                }
                catch (KlarnaException e)
                {
                    message = e.FaultString;
                    return false;
                }

                return true;
            }

            // Refund processing
            // when user click complete return order in commerce manager the transaction type will be Credit
            if (payment.TransactionType.Equals(TransactionType.Credit.ToString(), StringComparison.OrdinalIgnoreCase))
            {
                try
                {
                    // create a Klarna client
                    var klarnaProxy = KlarnaHelper.ConfigKlarna(new API(), PaymentType);
                    var returnAmount = payment.Amount;
                    var klarnaTransaction = payment.TransactionID;

                    string invoice = klarnaProxy.ReturnAmount(klarnaTransaction, (double)returnAmount, 0, API.Flag.IncVAT, string.Empty);
                    payment.TransactionID = invoice;
                }
                catch (KlarnaException e)
                {
                    message = e.FaultString;
                    return false;
                }

                return true;
            }

            // When "Complete" or "Refund" shipment in Commerce Manager, this method will be run again with the TransactionType is Capture/Credit
            // Klarna will always return true to bypass the payment process again.
            if (!payment.TransactionType.Equals(TransactionType.Authorization.ToString(), StringComparison.OrdinalIgnoreCase))
            {
                return true;
            }

            // Get the billing address
            var billingAddress = payment.Parent.Parent.OrderAddresses.Cast<OrderAddress>().FirstOrDefault(address2 => address2.Name == payment.BillingAddressId);
            if (billingAddress == null)
            {
                throw new PaymentException(PaymentException.ErrorType.ConfigurationError, "", "Billing address was not specified.");
            }

            // create a Klarna client
            var klarnaClient = KlarnaHelper.ConfigKlarna(new API(), PaymentType);

            // Add line items to pay by Klarna
            foreach (LineItem lineItem in payment.Parent.LineItems)
            {
                // Calculate the discount for each product in percent
                var discount = lineItem.ListPrice == 0 ? 0 : (double)(((lineItem.ListPrice * lineItem.Quantity - lineItem.ExtendedPrice) / lineItem.Quantity) / lineItem.ListPrice) * 100;
                klarnaClient.AddArticle((int)lineItem.Quantity, lineItem.LineItemId.ToString(), lineItem.DisplayName, (double)(lineItem.ListPrice), (double)KlarnaHelper.GetItemTax(lineItem, billingAddress), discount, API.GoodsIs.IncVAT);
            }

            // Get the cart
            var cart = payment.Parent.Parent as Cart;
            if (cart == null)
            {
                return false;
            }

            // Get the SocialSecurityNo from Session
            var socialSecurityNo = HttpContext.Current.Session[KlarnaHelper.SocialSecurityNo] as String;
            if (string.IsNullOrEmpty(socialSecurityNo))
            {
                return false;
            }

            var pclass = -1;
            var status = true;
            string houseNumber = string.Empty;
            string houseExtension = string.Empty;
            try
            {
                switch (billingAddress.CountryCode)
                {
                    case KlarnaHelper.Deu:
                        houseNumber = HttpContext.Current.Session[KlarnaHelper.GermanyHouseNumberText] as string;
                        // House number is required
                        if (string.IsNullOrEmpty(houseNumber))
                        {
                            status = false;
                            break;
                        }
                        break;
                    case KlarnaHelper.Nld:
                        houseNumber = HttpContext.Current.Session[KlarnaHelper.NetherlandsHouseNumberText] as string;
                        houseExtension = HttpContext.Current.Session[KlarnaHelper.NetherlandsHouseExtensionText] as string;
                        // House number is required
                        if (string.IsNullOrEmpty(houseNumber))
                        {
                            status = false;
                            break;
                        }
                        if (houseExtension == null)
                        {
                            houseExtension = string.Empty;
                        }

                        break;
                }

                // Add shipping fee and handling fee
                klarnaClient.AddArticle(1, string.Empty, "Shipment fee", (double) payment.Parent.ShippingTotal, 0, 0, API.GoodsIs.Shipping | API.GoodsIs.IncVAT);
                klarnaClient.AddArticle(1, string.Empty, "Handling fee", (double)payment.Parent.HandlingTotal, 0, 0, API.GoodsIs.Handling | API.GoodsIs.IncVAT);
                // Handles gift card and total amount of other active payment methods
                decimal otherActivePaymentAmount = payment.Parent.Total - payment.Amount;
                if (otherActivePaymentAmount != 0)
                {
                    // Klarna doesn't support Gift Card directly so we add special item with negative amount to make the actual amount correct
                    klarnaClient.AddArticle(1, string.Empty, "Gift Card", (double)-otherActivePaymentAmount, 0, 0, API.GoodsIs.Handling | API.GoodsIs.IncVAT);
                }

                KlarnaHelper.SetKlarnaAddress(klarnaClient, billingAddress, billingAddress, houseNumber, houseExtension);
                string[] result = klarnaClient.AddTransaction(socialSecurityNo, null, API.Flag.ReturnOCR, pclass);
                // save transaction id from Klarna server for later use
                payment.TransactionID = result[0];
            }
            catch (KlarnaException e)
            {
                _logger.Error("Error when processing Klarna invoice payment request", e);
                ErrorManager.GenerateError(e.FaultString);
                status = false;
            }
            return status;
        }

    }
}
