using System;
using System.Collections.Generic;
using System.Linq;
using System.Web.Script.Serialization;
using System.Web.UI;
using System.Web.UI.WebControls;
using EPiServer.Business.Commerce.Payment.Klarna.Services;
using EPiServer.Data.Dynamic;
using EPiServer.Framework.Localization;
using EPiServer.ServiceLocation;
using Klarna.Core;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Managers;
using Mediachase.Commerce.Website;
using Mediachase.Commerce.Website.BaseControls;
using Mediachase.Commerce.Website.Helpers;
using KlarnaAddress = Klarna.Core.Address;

namespace EPiServer.Business.Commerce.Payment.Klarna.KlarnaPartPayment
{
    /// <summary>
    ///	Klarna payment method
    /// </summary>
    public partial class PaymentMethod : BaseStoreUserControl, IPaymentOption
    {
        public const string UpdatedBillingAddress = "UpdatedBillingAddress";
        private CartHelper _currentCart;

        public PaymentMethod()
        {
            ErrorManager.Instance.Error += Instance_Error;
        }

        /// <summary>
        /// Handles the Error event of the Instance control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.IO.ErrorEventArgs"/> instance containing the event data.</param>
        private void Instance_Error(object sender, ErrorEventArgs e)
        {
            // Only show error if raise error from it self or from its payment gateway
            if (KlarnaHelper.IsCallingFrom(typeof(ErrorEventHandler), this.GetType().BaseType) || KlarnaHelper.IsCallingFrom(typeof(ErrorEventHandler), typeof(KlarnaPaymentGateway)))
            {
                ShowError(e.Message);
            }
        }


        /// <summary>
        /// Gets the cart of current user.
        /// </summary>
        /// <value>The <see cref="Mediachase.Commerce.Website.Helpers.CartHelper"/> cart of current user.</value>
        public CartHelper CurrentCart
        {
            get
            {
                if (_currentCart == null)
                {
                    _currentCart = new CartHelper(Cart.DefaultName);
                }

                return _currentCart;
            }
        }

        /// <summary>
        /// Gets the inputted billing address.
        /// </summary>
        /// <value>The <see cref="Mediachase.Commerce.Orders.OrderAddress"/> billing address.</value>
        public OrderAddress BillingAddress
        {
            get
            {
                return CurrentCart.FindAddressByName("Billing Address");
            }
        }

        /// <summary>
        /// Raises the <see cref="E:System.Web.UI.Control.Init"/> event.
        /// </summary>
        /// <param name="e">An <see cref="T:System.EventArgs"/> object that contains the event data.</param>
        protected override void OnInit(EventArgs e)
        {
            base.OnInit(e);
            ScriptManager.RegisterClientScriptInclude(Page, Page.GetType(), "KlarnaPartPaymentScript", "http://integration.klarna.com/js/klarnapart.js");
        }

        /// <summary>
        /// Handles the Load event of the Page control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void Page_Load(object sender, EventArgs e)
        {
            HideError();
            SelectAddressContainer.Visible = false;

            if (KlarnaHelper.IsUpdatePclass(KlarnaHelper.KlarnaPayment.PartPayment))
            {
                try
                {
                    // UpdatePclassValue
                    UpdatePclassValue(new API());
                }
                catch (KlarnaException klarnaEx)
                {
                    ErrorManager.GenerateError(string.Format("Update pclasses failed. {0}", klarnaEx.FaultString));
                    return;
                }
            }

        }

        #region IPaymentOption Members

        /// <summary>
        /// Validates input data for the control. In case of Credit card pre authentication will be the best way to
        /// validate. The error message if any should be displayed within a control itself.
        /// </summary>
        /// <returns>Returns false if validation is failed.</returns>
        public bool ValidateData()
        {
            // Hide the error container
            HideError();

            // It should be validate currency before validate address to show correct error message if has any.
            var isCurrencyValidate = KlarnaHelper.ValidateCurrency(BillingAddress.CountryCode);
            if (!isCurrencyValidate)
            {
                ShowError(
                    string.Format("{0} {1}", LocalizationService.GetString("/Commerce/klarnapartpayment/invalidcurrency"), 
                    KlarnaHelper.GetCurrencyNameByCode(KlarnaHelper.GetCurrencyByCountry(BillingAddress.CountryCode))));
            }
            return isCurrencyValidate && ValidateAddress();
        }

        /// <summary>
        /// Hides the error.
        /// </summary>
        private void HideError()
        {
            klarnaError.InnerText = string.Empty;
            klarnaErrorContainer.Visible = false;
        }

        /// <summary>
        /// Shows the error.
        /// </summary>
        /// <param name="errorMessage">The error message.</param>
        private void ShowError(string errorMessage)
        {
            klarnaError.InnerText = errorMessage;
            klarnaErrorContainer.Visible = true;
        }

        /// <summary>
        /// This method is called before the order is completed. This method should check all the parameters
        /// and validate the credit card or other parameters accepted.
        /// </summary>
        /// <param name="form">An OrderForm</param>
        /// <returns>A payment</returns>
        public Mediachase.Commerce.Orders.Payment PreProcess(OrderForm form)
        {
            if (!ValidateData())
            {
                // Intend to throw exception to stop processing unnecessary later steps (the sample site catch this exception to handle)
                throw new ArgumentNullException(
                    LocalizationService.GetString("/Commerce/klarnapartpayment/klarnadata"),
                    LocalizationService.GetString("/Commerce/klarnapartpayment/invaliddata"));
            }
            StoreInputDataToSession();

            // Create a new payment
            var otherPayment = new OtherPayment {BillingAddressId = form.BillingAddressId, TransactionType = TransactionType.Authorization.ToString()};

            // Get billing Address from session and update it if not null
            var billingAddress = Session[UpdatedBillingAddress] as Address;
            Session[UpdatedBillingAddress] = null;
            if (billingAddress != null)
            {
                KlarnaHelper.UpdateOrderAddress(form, billingAddress);
            }
            return otherPayment;
        }

        /// <summary>
        /// This method is called after the order is placed. This method should be used by the gateways that want to
        /// redirect customer to their site.
        /// </summary>
        /// <param name="form">The form.</param>
        /// <returns>True if success; otherwise false</returns>
        public bool PostProcess(OrderForm form)
        {
            var klarnaPaymentMethod = PaymentManager.GetPaymentMethodBySystemName(KlarnaHelper.KlarnaPartPayment, SiteContext.Current.LanguageName);
            var klarnaTranactionId = form.Payments.ToArray().FirstOrDefault(p => p.PaymentMethodId == klarnaPaymentMethod.PaymentMethod[0].PaymentMethodId).TransactionID;

            var klarnaClient = KlarnaHelper.ConfigKlarna(new API(), KlarnaHelper.KlarnaPayment.PartPayment);
            return KlarnaHelper.UpdateKlarnaInvoiceNumber(klarnaClient, form, klarnaTranactionId);
        }
        #endregion

        /// <summary>
        /// Deletes the store.
        /// </summary>
        /// <param name="name">The name.</param>
        private static void DeleteStore(string name)
        {
            DynamicDataStoreFactory.Instance.DeleteStore(name, true);
        }

        /// <summary>
        /// Updates the pclass value.
        /// </summary>
        /// <param name="klarnaClient">The klarna client.</param>
        private void UpdatePclassValue(API klarnaClient)
        {
            // Configure klarna client
            KlarnaConfig klarnaConfig = KlarnaHelper.GetKlarnaBaseConfig(KlarnaHelper.KlarnaPayment.PartPayment);
            klarnaConfig.PCStorage = KlarnaHelper.PcStorage;
            klarnaConfig.PCURI = KlarnaHelper.PcUriPartPayment;

            // Get the swedish campaigns and save them in DDS
            UpdatePclassByCountry(klarnaClient, klarnaConfig, KlarnaHelper.SwedishPclasssValues, KlarnaHelper.Swe);

            // Get the Finnish campaigns and save them in DDS
            UpdatePclassByCountry(klarnaClient, klarnaConfig, KlarnaHelper.FinnishPclasssValues, KlarnaHelper.Fin);

            // Get the German campaigns and save them in DDS
            UpdatePclassByCountry(klarnaClient, klarnaConfig, KlarnaHelper.GermanPclasssValues, KlarnaHelper.Deu);

            // Get the Norwegian campaigns and save them in DDS
            UpdatePclassByCountry(klarnaClient, klarnaConfig, KlarnaHelper.NorwegianPclasssValues, KlarnaHelper.Nor);

            // Get the Danish campaigns and save them in DDS
            UpdatePclassByCountry(klarnaClient, klarnaConfig, KlarnaHelper.DanishPclasssValues, KlarnaHelper.Dnk);

            // Get the Dutch campaigns and save them in DDS
            UpdatePclassByCountry(klarnaClient, klarnaConfig, KlarnaHelper.DutchPclasssValues, KlarnaHelper.Nld);

            // Get the Austrian campaigns and save them in DDS
            UpdatePclassByCountry(klarnaClient, klarnaConfig, KlarnaHelper.AustriaPclasssValues, KlarnaHelper.Aut);
        }

        /// <summary>
        /// Updates the pclass by country.
        /// </summary>
        /// <param name="klarnaClient">The klarna client.</param>
        /// <param name="config">The config.</param>
        /// <param name="storeName">Name of the store.</param>
        /// <param name="countryCode">The country code.</param>
        private void UpdatePclassByCountry(API klarnaClient, KlarnaConfig config, string storeName, string countryCode)
        {
            KlarnaHelper.UpdateKlarnaConfigByCountry(countryCode, config);
            klarnaClient.Config(config);
            klarnaClient.FetchPClasses();
            UpdatePclass(klarnaClient, storeName);
        }

        /// <summary>
        /// Updates the pclass.
        /// </summary>
        /// <param name="klarnaClient">The klarna client.</param>
        /// <param name="storeName">Name of the store.</param>
        private void UpdatePclass(API klarnaClient, string storeName)
        {
            // Delete the store
            DeleteStore(storeName);

            // Get the store
            var store = GetStore(storeName);
            List<PClass> pClasses = klarnaClient.GetPClasses(PClass.PClassType.NULL);
            foreach (PClass pclass in pClasses)
            {
                EPiPClass pc = new EPiPClass(pclass);
                // Save the pclassValue
                store.Save(pc);
            }
        }

        /// <summary>
        /// Gets the store.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns>Return a dynamic data store</returns>
        private static DynamicDataStore GetStore(string name)
        {
            return DynamicDataStoreFactory.Instance.GetStore(name) ??
                   DynamicDataStoreFactory.Instance.CreateStore(name, typeof(EPiPClass));
        }

        /// <summary>
        /// Gets the service URL of Klarna.
        /// </summary>
        /// <value>The service URL.</value>
        protected string ServicePath
        {
            get
            {
                return UriSupport.ResolveUrlBySettings("~/Services/KlarnaService.svc/");
            }
        }

        /// <summary>
        /// Gets the logo URL prefix to make the fully resourse url for displaying image logo for Klarna payment.
        /// </summary>
        /// <value>The logo URL prefix.</value>
        protected string LogoUrlPrefix
        {
            get
            {
                return ResolveUrl("~/Templates/Enoteca/Units/Placeable/plugins/payment/KlarnaPartPayment/ClientResources/klarna-");
            }
        }

        /// <summary>
        /// Validates the address.
        /// </summary>
        /// <returns>True if user input address correctly; otherwise false;</returns>
        private bool ValidateAddress()
        {
            var socialSecurityNo = PersonalNumberText.Text;

            switch (BillingAddress.CountryCode)
            {
                case KlarnaHelper.Swe:
                    if (string.IsNullOrEmpty(socialSecurityNo))
                    {
                        ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/selectsocialsecurityno"));
                        return false;
                    }

                    // If the enableCustomerEnterAddress is set, don't check for customer address
                    if (KlarnaHelper.IsAllowDifferentAddress(KlarnaHelper.KlarnaPayment.PartPayment))
                    {
                        return true;
                    }

                    // If user already select one candidate address.
                    if (SelectAddressRadioList.SelectedItem != null)
                    {
                        return true;
                    }

                    try
                    {
                        // Klarna client
                        var klarnaClient = KlarnaHelper.ConfigKlarna(new API(), KlarnaHelper.KlarnaPayment.PartPayment);
                        // Call Klarna api to get addresses
                        List<KlarnaAddress> klarnaAddresses = klarnaClient.GetAddresses(
                            socialSecurityNo,
                            KlarnaHelper.GetKlarnaEncodingByCountry(BillingAddress.CountryCode),
                            API.GetAddressFlag.GA_Last
                        );

                        var addresses = new List<Address>();
                        foreach (KlarnaAddress address in klarnaAddresses)
                        {
                            addresses.Add(new Address(address.Lastname, address.Street, address.ZipCode, address.City, address.Country));
                        }

                        // If multiple addresses for a company
                        if (addresses.Count > 1)
                        {
                            PopulateCompanyAddress(addresses);
                            ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/selectaddress"));
                            return false;
                        }

                        // If a single address, compare it with current billing address
                        var sameAddress = addresses[0].CompareWith(BillingAddress);
                        if (!sameAddress)
                        {
                            ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/invalidAddress"));
                        }
                        return sameAddress;
                    }
                    catch (KlarnaException ex)
                    {
                        // Show the error message
                        ShowError(ex.FaultString);
                        return false;
                    }

                case KlarnaHelper.Fin:
                case KlarnaHelper.Nor:
                case KlarnaHelper.Dnk:
                    if (string.IsNullOrEmpty(socialSecurityNo))
                    {
                        ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/selectsocialsecurityno"));
                        return false;
                    }

                    return true;
                case KlarnaHelper.Deu:
                    if (string.IsNullOrEmpty(GermanyBirthDateText.Text))
                    {
                        ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/selectbirthdate"));
                        return false;
                    }

                    if (string.IsNullOrEmpty(GermanyHouseNumberText.Text))
                    {
                        ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/selecthousenumber"));
                        return false;
                    }

                    return true;
                case KlarnaHelper.Nld:
                    if (string.IsNullOrEmpty(NetherlandsBirthDateText.Text))
                    {
                        ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/selectbirthdate"));
                        return false;
                    }

                    if (string.IsNullOrEmpty(NetherlandsHouseNumberText.Text))
                    {
                        ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/selecthousenumber"));
                        return false;
                    }

                    return true;
                case KlarnaHelper.Aut:
                    if (string.IsNullOrEmpty(AustriaBirthDateText.Text))
                    {
                        ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/selectbirthdate"));
                        return false;
                    }

                    return true;
                default:
                    // Klarna module does not support other countries
                    ShowError(LocalizationService.GetString("/Commerce/klarnapartpayment/klarnanotavailable"));
                    return false;
            }
        }

        /// <summary>
        /// Stores the input data to session.
        /// </summary>
        private void StoreInputDataToSession()
        {
            // Get the SocialSecurityNo and add to Session
            var socialSecurityNo = PersonalNumberText.Text;

            switch (BillingAddress.CountryCode)
            {
                case KlarnaHelper.Swe:
                    Session[KlarnaHelper.SocialSecurityNo] = socialSecurityNo;
                    if (SelectAddressRadioList.SelectedItem != null)
                    {
                        Session[UpdatedBillingAddress] = new JavaScriptSerializer().Deserialize<Address>(SelectAddressRadioList.SelectedValue);
                    }
                    break;

                case KlarnaHelper.Fin:
                case KlarnaHelper.Nor:
                case KlarnaHelper.Dnk:
                    Session[KlarnaHelper.SocialSecurityNo] = socialSecurityNo;
                    break;

                case KlarnaHelper.Deu:
                    Session[KlarnaHelper.GermanyHouseNumberText] = GermanyHouseNumberText.Text;
                    Session[KlarnaHelper.SocialSecurityNo] = string.Format("{0}{1}", GermanyBirthDateText.Text, GermanyGenderMale.Checked ? "1" : "0");
                    break;
                case KlarnaHelper.Nld:
                    Session[KlarnaHelper.NetherlandsHouseNumberText] = NetherlandsHouseNumberText.Text;
                    Session[KlarnaHelper.NetherlandsHouseExtensionText] = NetherlandsHouseExtensionText.Text;
                    Session[KlarnaHelper.SocialSecurityNo] = string.Format("{0}{1}", NetherlandsBirthDateText.Text, NetherlandsGenderMale.Checked ? "1" : "0");
                    break;
                case KlarnaHelper.Aut:
                    Session[KlarnaHelper.SocialSecurityNo] = string.Format("{0}{1}", AustriaBirthDateText.Text, AustriaGenderMale.Checked ? "1" : "0");
                    break;
            }

            // Save the pclassId to Session
            Session[KlarnaPaymentGateway.PclassId] = UpdatedPclassIdValue.Value;
        }

        /// <summary>
        /// Populates the company address.
        /// </summary>
        /// <param name="addresses">The addresses.</param>
        private void PopulateCompanyAddress(List<Address> addresses)
        {
            // Please choose the address below to be used as delivery and invoice address. The delivery and 
            // invoice address on your order will be updated with the chosen address.
            SelectAddressContainer.Visible = true;
            SelectAddressRadioList.Items.Clear();
            var serializer = new JavaScriptSerializer();
            foreach (var companyAddress in addresses)
            {
                var item = new ListItem(companyAddress.ToString(), serializer.Serialize(companyAddress));
                item.Attributes.Add("style", "padding-left: 20px; padding-top: 7px; display: block;");
                SelectAddressRadioList.Items.Add(item);
            }
        }

        private static LocalizationService LocalizationService
        {
            get { return ServiceLocator.Current.GetInstance<LocalizationService>(); }
        }
    }
}
