using System;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Collections;
using System.Drawing;
using System.Workflow.ComponentModel.Compiler;
using System.Workflow.ComponentModel.Serialization;
using System.Workflow.ComponentModel;
using System.Workflow.ComponentModel.Design;
using System.Workflow.Runtime;
using System.Workflow.Activities;
using System.Workflow.Activities.Rules;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Marketing;
using Mediachase.Commerce.Marketing.Objects;
using System.Collections.Generic;
using Mediachase.Commerce.Catalog;
using Mediachase.Commerce.Catalog.Dto;
using Mediachase.Commerce.Customers.Profile;
using System.Collections.Specialized;
using Mediachase.Commerce.Catalog.Managers;
using System.Web.Security;
using System.Linq;

namespace Mediachase.Commerce.Workflow.Activities
{
	public partial class CheckInventoryActivity : OrderGroupActivityBase
	{
		/// <summary>
		/// Initializes a new instance of the <see cref="CheckInventoryActivity"/> class.
		/// </summary>
		public CheckInventoryActivity()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Called by the workflow runtime to execute an activity.
		/// </summary>
		/// <param name="executionContext">The <see cref="T:System.Workflow.ComponentModel.ActivityExecutionContext"/> to associate with this <see cref="T:System.Workflow.ComponentModel.Activity"/> and execution.</param>
		/// <returns>
		/// The <see cref="T:System.Workflow.ComponentModel.ActivityExecutionStatus"/> of the run task, which determines whether the activity remains in the executing state, or transitions to the closed state.
		/// </returns>
		protected override ActivityExecutionStatus Execute(ActivityExecutionContext executionContext)
		{
			try
			{
				// Validate the properties at runtime
				this.ValidateRuntime();

				// Calculate order discounts
				this.ValidateItems();

				// Retun the closed status indicating that this activity is complete.
				return ActivityExecutionStatus.Closed;
			}
			catch
			{
				// An unhandled exception occured.  Throw it back to the WorkflowRuntime.
				throw;
			}
		}

		private void ValidateItems()
		{
			//We don't need to validate quantity in the wishlist
			var orderForms = OrderGroup.OrderForms.ToArray();
			var lineItems = orderForms.SelectMany(x => x.LineItems.ToArray());
			var validLineItems = lineItems.Where(x => x.CatalogEntryId != "0" && !String.IsNullOrEmpty(x.CatalogEntryId) && !x.CatalogEntryId.StartsWith("@"));
			foreach (LineItem lineItem in validLineItems)
			{
				List<string> changeQtyReason = new List<string>();
				var newQty = GetNewLineItemQty(lineItem, changeQtyReason);
				var changeQtyReasonDisplay = String.Join(" and ", changeQtyReason.ToArray());
				if (newQty == 0)
				{
					// Remove item if it reached this stage
					Warnings.Add("LineItemRemoved-" + lineItem.Id.ToString(), String.Format("Item \"{0}\" has been removed from the cart because it is no longer available.", lineItem.DisplayName));
					DeleteLineItemFromShipments(lineItem);
					// Delete item
					lineItem.Delete();
				}
				else
				{
					var delta = lineItem.Quantity - newQty;
					if (delta != 0)
					{
						lineItem.Quantity -= delta;
						ChangeShipmentsLineItemQty(lineItem, delta);
						Warnings.Add("LineItemQtyChanged-" + lineItem.Id.ToString(),
									 String.Format("Item \"{0}\" quantity has been changed {1}.", lineItem.DisplayName, changeQtyReasonDisplay));
					}
				}
			}

		}

		private decimal GetNewLineItemQty(LineItem lineItem, List<string> changeQtyReason)
		{
			var newLineItemQty = lineItem.Quantity;

			if (newLineItemQty < lineItem.MinQuantity || newLineItemQty > lineItem.MaxQuantity)
			{
				newLineItemQty = Math.Max(lineItem.MinQuantity, newLineItemQty);
				if (newLineItemQty != lineItem.Quantity)
				{
					changeQtyReason.Add("by Min Quantity setting");
				}
				newLineItemQty = Math.Min(lineItem.MaxQuantity, newLineItemQty);
				if (newLineItemQty != lineItem.Quantity)
				{
					changeQtyReason.Add("by Max Quantity setting");
				}
			}

			if (lineItem.InventoryStatus == InventoryStatus.Enabled)
			{
				// Check Inventory
				// item exists with appropriate quantity
				if (lineItem.InStockQuantity < newLineItemQty)
				{
					if (lineItem.InStockQuantity > 0) // there still exist items in stock
					{
						// check if we can backorder some items
						if (lineItem.AllowBackordersAndPreorders)
						{
							//Increase stock qty by backorder qty if 
							var availStockAndBackorderQty = lineItem.InStockQuantity + lineItem.BackorderQuantity;
							if (availStockAndBackorderQty >= newLineItemQty)
							{
								//NONE
							}
							else
							{
								newLineItemQty = availStockAndBackorderQty;
								changeQtyReason.Add("by BackOrder quantity");
							}
						}
						else
						{
							newLineItemQty = lineItem.InStockQuantity;
						}
					}
					else
					{
						if (lineItem.AllowBackordersAndPreorders && lineItem.PreorderQuantity > 0)
						{
							if (lineItem.PreorderQuantity >= newLineItemQty)
							{
								//NONE
							}
							else
							{
								newLineItemQty = lineItem.PreorderQuantity;
								changeQtyReason.Add("by Preorder quantity");
							}
						}
						else if (lineItem.AllowBackordersAndPreorders && lineItem.BackorderQuantity > 0)
						{
							if (lineItem.BackorderQuantity >= newLineItemQty)
							{
							}
							else
							{
								newLineItemQty = lineItem.BackorderQuantity;
								changeQtyReason.Add("by BackOrder quantity");
							}
						}
						else
						{
							newLineItemQty = 0;
						}
					}
				}
			}
			return newLineItemQty;
		}

		private void DeleteLineItemFromShipments(LineItem lineItem)
		{
			var orderForm = OrderGroup.OrderForms.ToArray().FirstOrDefault();
			if (orderForm != null)
			{
				var lineItemIndex = orderForm.LineItems.IndexOf(lineItem);
				var allShipmentContainsLineItem = orderForm.Shipments.ToArray().Where(x => Shipment.GetShipmentLineItems(x).Contains(lineItem));
				foreach (var shipment in allShipmentContainsLineItem)
				{
					shipment.RemoveLineItemIndex(lineItemIndex);
				}
			}
		}
		private void ChangeShipmentsLineItemQty(LineItem lineItem, decimal delta)
		{
			var orderForm = OrderGroup.OrderForms.ToArray().FirstOrDefault();
			if (orderForm != null)
			{
				var lineItemIndex = orderForm.LineItems.IndexOf(lineItem);
				var allShipmentContainsLineItem = orderForm.Shipments.ToArray().Where(x => Shipment.GetShipmentLineItems(x).Contains(lineItem));
				foreach (var shipment in allShipmentContainsLineItem)
				{
					//Decrease qty in all shipment contains line item
					var shipmentQty = Shipment.GetLineItemQuantity(shipment, lineItem.LineItemId);
					var newShipmentQty = shipmentQty - delta;
					newShipmentQty = newShipmentQty > 0 ? newShipmentQty : 0;
					//Set new line item qty in shipment
					shipment.SetLineItemQuantity(lineItemIndex, newShipmentQty);
					delta -= Math.Min(delta, shipmentQty);

					if (delta == 0)
						break;
				}
			}
		}
	}
}
