﻿using System;
using System.Collections.Specialized;
using System.IO;
using System.Net;
using System.Text;
using System.Web;
using EPiServer.Core;
using log4net;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Dto;
using Mediachase.Commerce.Orders.Managers;
using Mediachase.Commerce.Website;

namespace EPiServer.Business.Commerce.Payment.PayPal
{
    public partial class PayPal : System.Web.UI.Page
    {
        private static ILog _logger = LogManager.GetLogger(typeof(PayPal));

        /// <summary>
        /// Raises the <see cref="E:System.Web.UI.Control.Load"/> event.
        /// </summary>
        /// <param name="e">The <see cref="T:System.EventArgs"/> object that contains the event data.</param>
        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            Providers.StaticUrlProvider urls = new Providers.StaticUrlProvider();
            urls.Initialize("UrlProvider", new NameValueCollection());

            // Get PaypalPaymentLanding url
            var referencePage = DataFactory.Instance.GetPage(PageReference.StartPage)["PaypalPaymentLandingPage"] as PageReference;
            var landingPage = DataFactory.Instance.GetPage(referencePage);
            string acceptUrl = landingPage != null ? landingPage.LinkURL : string.Empty;

            string cancelUrl = Utilities.GetUrlValueFromStartPage("CheckoutPage") ?? string.Empty;  // get link to ShoppingCart
            cancelUrl += "#state=step|2&success=false&paymentmethod=paypal";
            string orderID = string.Empty, paypalExpToken = string.Empty, isoFilePath = string.Empty;
            Utilities.GetValuesFromCookies(ref orderID, ref paypalExpToken, ref isoFilePath);
            

            if(!string.IsNullOrEmpty(Request.QueryString["hash"]) && string.Equals(Request.QueryString["accept"], "true"))
            {
                PayPalPaymentGateway gateway = new PayPalPaymentGateway();
                if (!string.IsNullOrEmpty(orderID) && PayPalPaymentGateway.GetMD5Key(orderID + "accepted") == Request.QueryString["hash"])
                {
                    gateway.ProcessSuccessfulTransaction(acceptUrl, cancelUrl);
                }
            }
            else if (!string.IsNullOrEmpty(Request.QueryString["hash"]) && string.Equals(Request.QueryString["accept"], "false"))
            {
                PayPalPaymentGateway gateway = new PayPalPaymentGateway();
                if (!string.IsNullOrEmpty(orderID) && PayPalPaymentGateway.GetMD5Key(orderID + "canceled") == Request.QueryString["hash"])
                {
                    gateway.ProcessUnsuccessfulTransaction(cancelUrl, Utilities.Translate("PayPalCancelMessage"));
                }
            }
            else if (Request.Form["receiver_email"] != null && Request.Form["txn_id"] != null)
            {
                // if Request contain param "receiver_email" and "txn_id", it means this is IPN message from PayPal.com
                ProcessIPNMessageFromPayPal();
            }

            
        }

        

        /// <summary>
        /// This function processes Instant Payment Notification (IPN) from PayPal. 
        /// About how to process this message, see https://cms.paypal.com/cgi-bin/marketingweb?cmd=_render-content&content_ID=developer/e_howto_admin_IPNIntro
        /// <remarks>
        /// IPN is PayPal's message service that sends a notification when a transaction is affected.
        /// </remarks>
        /// </summary>
        private void ProcessIPNMessageFromPayPal()
        {
            PaymentMethodDto payPal = PaymentManager.GetPaymentMethodBySystemName("PayPal", SiteContext.Current.LanguageName);
            string userName = payPal.GetParameterValueByName(PayPalPaymentGateway.BusinessEmailParameter);
            
            // username should be equal with receiver_email
            if (!string.Equals(userName, Request.Form["receiver_email"] as string, StringComparison.OrdinalIgnoreCase))
            {
                return;
            }

            #region Validate back to know that this request is really come from PayPal.com

            string url = payPal.GetParameterValueByName(PayPalPaymentGateway.ExpChkoutURLParameter);
            byte[] param = Request.BinaryRead(HttpContext.Current.Request.ContentLength);
            string strRequest = Encoding.ASCII.GetString(param) + "&cmd=_notify-validate";

            // Request content will be the whole IPN message's url
            HttpWebRequest req = (HttpWebRequest)WebRequest.Create(url);
            req.Method = "POST";
            req.ContentType = "application/x-www-form-urlencoded";
            req.ContentLength = strRequest.Length;
            StreamWriter streamOut = new StreamWriter(req.GetRequestStream(), System.Text.Encoding.ASCII);
            streamOut.Write(strRequest);
            streamOut.Close();

            //Send the request to PayPal and get the response
            StreamReader streamIn = new StreamReader(req.GetResponse().GetResponseStream());
            string strResponse = streamIn.ReadToEnd();
            streamIn.Close();

            if (strResponse.Equals("INVALID", StringComparison.OrdinalIgnoreCase))
            {
                _logger.Error(string.Format("INVALID IPN message received from PayPal, original message: {0}", strRequest));
                return;
            }
            else if (!strResponse.Equals("VERIFIED", StringComparison.OrdinalIgnoreCase))
            {
                _logger.Error(string.Format("UNKNOWN IPN message received from PayPal, original message: {0}", strRequest));
                return;
            }

            // NOTE: we only process the VERIFIED IPN message from PayPal

            #endregion


            string custom = Request.Form["custom"];
            string status = Request.Form["payment_status"];
            string reason = Request.Form["pending_reason"];
            string payAction = payPal.GetParameterValueByName(PayPalPaymentGateway.PaymentActionParameter);
            if (payAction == "Authorization" && string.Equals(status, "pending", StringComparison.OrdinalIgnoreCase))
            {
                if (string.Equals(reason, "authorization", StringComparison.OrdinalIgnoreCase))
                {
                    //Update the order status here
                    try
                    {
                        string[] values = custom.Split('|');
                        PurchaseOrder[] orders = PurchaseOrder.LoadByCustomer(new Guid(values[0])).ToArray();
                        foreach (PurchaseOrder po in orders)
                        {
                            if (po.TrackingNumber == values[1])
                            {
                                po.Status = "Processing";
                                po.AcceptChanges();
                                break;
                            }
                        }
                    }
                    catch (System.Exception ex)
                    {
                        _logger.Error(string.Format("Error while processing message received from PayPal: {0}", ex.Message));
                    }
                }
                else
                {
                    _logger.Warn(string.Format("Payment is pending at PayPal with following reason: {0}", reason));
                    //Update the order status here
                    try
                    {
                        string[] values = custom.Split('|');
                        PurchaseOrder[] orders = PurchaseOrder.LoadByCustomer(new Guid(values[0])).ToArray();
                        foreach (PurchaseOrder po in orders)
                        {
                            if (po.TrackingNumber == values[1])
                            {
                                po.Status = "Submitted";
                                po.AcceptChanges();
                                break;
                            }
                        }
                    }
                    catch (System.Exception ex)
                    {
                        _logger.Error(string.Format("Error while processing message received from PayPal: {0}", ex.Message));
                    }
                }
            }
            else if (payAction == "Sale")
            {
                if (string.Equals(status, "completed", StringComparison.OrdinalIgnoreCase)
                    || string.Equals(status, "processed", StringComparison.OrdinalIgnoreCase))
                {
                    //Update the order status here
                    try
                    {
                        string[] values = custom.Split('|');
                        PurchaseOrder[] orders = PurchaseOrder.LoadByCustomer(new Guid(values[0])).ToArray();
                        foreach (PurchaseOrder po in orders)
                        {
                            if (po.TrackingNumber == values[1])
                            {
                                po.Status = "Shipped";
                                po.AcceptChanges();
                                break;
                            }
                        }
                    }
                    catch (System.Exception ex)
                    {
                        _logger.Error(string.Format("Error while processing message received from PayPal: {0}", ex.Message));
                    }
                }
                else
                {
                    _logger.Warn(string.Format("Payment is not successful at PayPal with following status code: {0}", status));
                    //Update the order status here
                    try
                    {
                        string[] values = custom.Split('|');
                        PurchaseOrder[] orders = PurchaseOrder.LoadByCustomer(new Guid(values[0])).ToArray();
                        foreach (PurchaseOrder po in orders)
                        {
                            if (po.TrackingNumber == values[1])
                            {
                                po.Status = "Submitted";
                                po.AcceptChanges();
                                break;
                            }
                        }
                    }
                    catch (System.Exception ex)
                    {
                        _logger.Error(string.Format("Error while processing message received from PayPal: {0}", ex.Message));
                    }
                }

            }
            else
            {
                _logger.Error(string.Format("Invalid status and status reason received from PayPal, original message: {0}", strRequest));
            }

        }

        
    }
}
