﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Mediachase.Search.Extensions;
using Mediachase.Commerce;

namespace Mediachase.Search.Providers.Solr35.Select
{
    internal class FacetManager : IFacetManager
    {
        #region Facet Count Response Wrappers

        public class FieldResponse : IFacetFieldResponse
        {
            public string Field { get; private set; }
            public string Value { get; private set; }
            public int Count { get; private set; }
            public FieldResponse(string field, string value, int count)
            {
                this.Field = field;
                this.Value = value;
                this.Count = count;
            }
        }

        public class QueryResponse : IFacetQueryResponse
        {
            public string Query { get; private set; }
            public int Count { get; private set; }
            public QueryResponse(string query, int count)
            {
                this.Query = query;
                this.Count = count;
            }
        }

        #endregion


        #region Private Implementation

        private class FieldRequest
        {
            public string Field { get; private set; }
            public SearchFilter Filter { get; private set; }
            public Dictionary<string, SimpleValue> ValueMap { get; private set; }

            public FacetGroup ResultGroup { get; set; }

            public FieldRequest(SearchFilter filter)
            {
                this.Field = filter.field.ToLowerInvariant();
                this.Filter = filter;
                this.ValueMap = filter.Values.SimpleValue.ToDictionary(sv => sv.value.ToLowerInvariant());
            }
        }

        private class QueryRequestGroup
        {
            public string Field { get; private set; }
            public SearchFilter Filter { get; private set; }
            public FacetGroup ResultGroup { get; set; }

            public QueryRequestGroup(SearchFilter filter)
            {
                this.Field = filter.field.ToLowerInvariant();
                this.Filter = filter;
            }
        }

        private class QueryRequest
        {
            public QueryRequestGroup Group { get; private set; }
            public string Query { get; private set; }
            public string Key { get; private set; }
            public Descriptions Descriptions { get; private set; }

            public QueryRequest(QueryRequestGroup group, RangeValue value)
            {
                this.Group = group;
                this.Query = new RangeQuery(group.Field, 
                    value.lowerbound, value.upperbound, 
                    value.lowerboundincluded, value.upperboundincluded).ToString();
                this.Key = value.key;
                this.Descriptions = value.Descriptions;
            }

            public QueryRequest(QueryRequestGroup group, PriceRangeValue value, Currency currency)
            {
                this.Group = group;
                this.Query = new RangeQuery(string.Format("{0}{1}", group.Field, currency.CurrencyCode.ToLowerInvariant()),
                    value.lowerbound, value.upperbound, 
                    value.lowerboundincluded, value.upperboundincluded).ToString();
                this.Key = value.key;
                this.Descriptions = value.Descriptions;
            }

            // removed for caompatibility release.
            //public QueryRequest(QueryRequestGroup group, PriceRangeValue value, Currency currency, MarketId marketId)
            //{
            //    this.Group = group;
            //    this.Query = new RangeQuery(string.Format("{0}{1}_{2}", group.Field, currency.CurrencyCode.ToLowerInvariant(), marketId.Value.ToLowerInvariant()),
            //        value.lowerbound, value.upperbound,
            //        value.lowerboundincluded, value.upperboundincluded).ToString();
            //    this.Key = value.key;
            //    this.Descriptions = value.Descriptions;
            //}
        }

        private Dictionary<string, List<FieldRequest>> fieldRequestMap;
        private Dictionary<string, List<QueryRequest>> queryRequestMap;
        private List<QueryRequestGroup> queryRequestGroups;

        private void AddRequest(FieldRequest fieldRequest)
        {
            List<FieldRequest> requestList;
            if (fieldRequestMap.TryGetValue(fieldRequest.Field, out requestList)) requestList.Add(fieldRequest);
            else
            {
                requestList = new List<FieldRequest>();
                requestList.Add(fieldRequest);
                fieldRequestMap.Add(fieldRequest.Field, requestList);
            }
        }

        private void AddRequest(QueryRequest queryRequest)
        {
            List<QueryRequest> requestList;
            if (queryRequestMap.TryGetValue(queryRequest.Query, out requestList)) requestList.Add(queryRequest);
            else
            {
                requestList = new List<QueryRequest>();
                requestList.Add(queryRequest);
                queryRequestMap.Add(queryRequest.Query, requestList);
            }
        }

        #endregion


        public IEnumerable<string> FacetFields { get { return fieldRequestMap.Keys; } }

        public IEnumerable<string> FacetQueries { get { return queryRequestMap.Keys; } }

        public FacetManager(IEnumerable<SearchFilter> facetRequests, Currency currency)
        {
            // market id parameter removed for compatibility release.

            fieldRequestMap = new Dictionary<string, List<FieldRequest>>();
            queryRequestMap = new Dictionary<string, List<QueryRequest>>();
            queryRequestGroups = new List<QueryRequestGroup>();

            foreach (SearchFilter filter in facetRequests)
            {
                bool isFieldRequest = filter.Values.SimpleValue != null && filter.Values.SimpleValue.Length > 0;
                bool isRangeRequest = filter.Values.RangeValue != null && filter.Values.RangeValue.Length > 0;
                bool isPriceRequest = filter.Values.PriceRangeValue != null &&
                    filter.Values.PriceRangeValue.Any(prv => prv.currency.Equals(currency, StringComparison.OrdinalIgnoreCase));

                int requestTypeCount = (isFieldRequest ? 1 : 0) + (isRangeRequest ? 1 : 0) + (isPriceRequest ? 1 : 0);
                if (requestTypeCount == 1)
                {
                    if (isFieldRequest) AddRequest(new FieldRequest(filter));
                    else
                    {
                        QueryRequestGroup requestGroup = new QueryRequestGroup(filter);
                        queryRequestGroups.Add(requestGroup);
                        if (isRangeRequest)
                        {
                            foreach (RangeValue value in filter.Values.RangeValue)
                            {
                                AddRequest(new QueryRequest(requestGroup, value));
                            }
                        }
                        else
                        {
                            foreach (PriceRangeValue value in filter.Values.PriceRangeValue
                                .Where(prv => string.Equals(prv.currency, currency.CurrencyCode, StringComparison.OrdinalIgnoreCase)))
                            {
                                // altered for compatibility release.
                                //if (string.IsNullOrEmpty(value.market) && string.IsNullOrEmpty(marketId))
                                //{
                                    AddRequest(new QueryRequest(requestGroup, value, currency));
                                //}
                                //else if (!string.IsNullOrEmpty(value.market) && string.Equals(value.market, marketId.Value, StringComparison.OrdinalIgnoreCase))
                                //{
                                //    AddRequest(new QueryRequest(requestGroup, value, currency, marketId));
                                //}
                            }
                        }
                    }
                }
            }
        }

        public IEnumerable<ISearchFacetGroup> BuildResponse(IEnumerable<IFacetFieldResponse> fieldCounts, IEnumerable<IFacetQueryResponse> queryCounts, string locale, bool facetLocalizedFieldValuesOnly)
        {
            return BuildFieldResponse(fieldCounts, locale, facetLocalizedFieldValuesOnly)
                .Concat(BuildQueryResponse(queryCounts, locale));
        }

        private IEnumerable<ISearchFacetGroup> BuildFieldResponse(IEnumerable<IFacetFieldResponse> fieldCounts, string locale, bool specifiedValuesOnly)
        {            
            foreach (FieldResponse response in fieldCounts
                .Where(resp => !string.IsNullOrEmpty(resp.Value))) // exclude facets for blank values.
            {
                List<FieldRequest> requests;
                if (fieldRequestMap.TryGetValue(response.Field, out requests))
                {
                    foreach (FieldRequest request in requests)
                    {
                        if (request.ResultGroup == null)
                        {
                            string localizedGroupName = SolrSearchProvider.GetLocalizedDescription(request.Filter.Descriptions, locale);
                            request.ResultGroup = new FacetGroup(response.Field, localizedGroupName);
                        }

                        SimpleValue value;
                        if (request.ValueMap.TryGetValue(response.Value.ToLowerInvariant(), out value))
                        {
                            string localizedValueName = SolrSearchProvider.GetLocalizedDescription(value.Descriptions, locale);
                            Facet facet = new Facet(request.ResultGroup, value.key, localizedValueName, response.Count);
                            request.ResultGroup.Facets.Add(facet);
                        }
                        else if (!specifiedValuesOnly)
                        {
                            Facet facet = new Facet(request.ResultGroup, null, response.Value, response.Count);
                            request.ResultGroup.Facets.Add(facet);
                        }
                    }
                }
            }

            foreach (FieldRequest fieldRequest in fieldRequestMap.Values.SelectMany(x => x))
            {
                if (fieldRequest.ResultGroup != null) yield return fieldRequest.ResultGroup;
            }
        }

        private IEnumerable<ISearchFacetGroup> BuildQueryResponse(IEnumerable<IFacetQueryResponse> queryCounts, string locale)
        {
            foreach (QueryResponse response in queryCounts)
            {
                List<QueryRequest> requests;
                if (queryRequestMap.TryGetValue(response.Query, out requests))
                {
                    foreach (QueryRequest request in requests)
                    {
                        if (request.Group.ResultGroup == null)
                        {
                            string localizedGroupName = SolrSearchProvider.GetLocalizedDescription(request.Group.Filter.Descriptions, locale);
                            request.Group.ResultGroup = new FacetGroup(request.Group.Filter.field, localizedGroupName);
                        }

                        string localizedQueryName = SolrSearchProvider.GetLocalizedDescription(request.Descriptions, locale);
                        Facet facet = new Facet(request.Group.ResultGroup, request.Key, localizedQueryName, response.Count);
                        request.Group.ResultGroup.Facets.Add(facet);
                    }
                }
            }

            foreach (QueryRequestGroup queryRequestGroup in queryRequestGroups)
            {
                if (queryRequestGroup.ResultGroup != null) yield return queryRequestGroup.ResultGroup;
            }
        }
    }
}
