﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web;
using System.Globalization;
using System.Threading;
using System.Collections.Specialized;
using Mediachase.Commerce.Shared;

namespace Mediachase.Commerce.Website
{
    /// <summary>
    /// Contain information about current site context for the current user. The object does not store any values between requests.
    /// You must make sure you initialize all custom values before they are used or default values will be returned.
    /// </summary>
    public class SiteContext
    {
        private static readonly object _lockObject = new object();
        private Guid _SiteId;

        public static string SITECONTEXT = "Mediachase.SiteContext";
		public static string CURRENCYCOOKIE = "currencyCode";

        /// <summary>
        /// Initializes a new instance of the <see cref="SiteContext"/> class.
        /// </summary>
        private SiteContext()
        { 
        }

        /// <summary>
        /// Gets the current.
        /// </summary>
        /// <value>The current.</value>
        public static SiteContext Current
        {
            get
            {
                HttpContext context = HttpContext.Current;
                if (context == null)
                    return new SiteContext();

                if (context.Items[SITECONTEXT] == null)
                {
                    lock (_lockObject)
                    {
                        if (context.Items[SITECONTEXT] == null)
                        {
                            SiteContext siteContext = new SiteContext();
                            if (!context.Items.Contains(SITECONTEXT))
                                context.Items.Add(SITECONTEXT, siteContext);
                            else
                                context.Items[SITECONTEXT] = siteContext;
                        }
                    }
                }

                return (SiteContext)context.Items[SITECONTEXT];
            }
        }

        private Currency _Currency = String.Empty;
        /// <summary>
        /// Gets or sets the currency code. Currency code is automatically taken from the CultureInfo.CurrentCulture property. Which is a
        /// culture for a current thread. The property can be changed by setting the currency directly.
        /// </summary>
        /// <value>The currency.</value>
        public Currency Currency
        {
            get
            {
                if (String.IsNullOrEmpty(_Currency))
                {
					NameValueCollection cookies = CommonHelper.GetCookie(CURRENCYCOOKIE);

                    if (cookies != null && cookies.Count > 0)
                        _Currency = cookies[0];
                    else
                        _Currency = CurrencyFormatter.CurrentCurrency();
                }

                return _Currency;
            }
            set
            {
                _Currency = value;
				CommonHelper.SetCookie(CURRENCYCOOKIE, value.CurrencyCode, DateTime.Now.AddDays(30));
            }
        }

        private string _CurrencyCode = String.Empty;
        /// <summary>
        /// Gets or sets the currency code. Currency code is automatically taken from the CultureInfo.CurrentCulture property. Which is a
        /// culture for a current thread. The property can be changed by setting the currency directly.
        /// </summary>
        /// <value>The currency code.</value>
        [Obsolete("Use the Currency member instead.")]
        public string CurrencyCode
        {
            get
            {
                return Currency;
            }
            set
            {
                Currency = new Currency(value);
            }
        }

        /// <summary>
        /// Gets or sets the site id that current CMS system uses.
        /// </summary>
        /// <value>The site id.</value>
        public Guid SiteId
        {
            get
            {
                return _SiteId;
            }
            set
            {
                _SiteId = value;
            }
        }

        private string _appPath = null;

        /// <summary>
        /// Gets or sets the app path.
        /// </summary>
        /// <value>The app path.</value>
        public string AppPath
        {
            get
            {
                return _appPath;
            }
            set
            {
                _appPath = value;
            }
        }

        private string _languageName;


        /// <summary>
        /// Gets or sets the name of the language.
        /// </summary>
        /// <value>The name of the language.</value>
        public string LanguageName
        {
            get
            {
                if (!String.IsNullOrEmpty(_languageName))
                    return _languageName;

                if (String.IsNullOrEmpty(_languageName))
                    _languageName = Thread.CurrentThread.CurrentCulture.Name;

                if (String.IsNullOrEmpty(_languageName))
                    _languageName = "en-US";

                return _languageName;
            }
            set
            {
                _languageName = value;
            }
        }

        /// <summary>
        /// Returns a <see cref="System.String"/> that represents this instance.
        /// </summary>
        /// <returns>
        /// A <see cref="System.String"/> that represents this instance.
        /// </returns>
        public override string ToString()
        {
            return String.Format("{0} - {1} - {2}", this.AppPath, this.LanguageName, this.SiteId.ToString());
        }
    }
}
