﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using Mediachase.Search.Extensions;
using Mediachase.Search.Providers.Solr35.Select;

namespace Mediachase.Search.Providers.Solr35
{
    public class SolrSearchQueryBuilder : ISearchQueryBuilder
    {
        // matches character sequences -, +, !, ^, [, ], {, }, (, ), ", !, :, ;, \, ?, &&, and ||; and matches * unless it is the last non-whitespace character.
        private static readonly Regex SpecialCharacters = new Regex(@"([-[+!^{}()""~:;\\?]|\]|&&|\|\||\*[\s]*(?=[^\s]))", RegexOptions.Compiled);
        
        private static readonly string MinDateTimeString = DateTime.MinValue.ToString(SolrSearchProvider.DateTimeFormatString);
        private static readonly string MaxDateTimeString = DateTime.MaxValue.ToString(SolrSearchProvider.DateTimeFormatString);


        public static string EscapeValueString(string value)
        {            
            string result = SpecialCharacters.Replace(value, "\\$1");
            return result.Contains(' ') ? string.Format("\"{0}\"", result) : result;
        }


        private class SolrSearchQuery : ISolrSearchQuery
        {
            private List<KeyValuePair<string, string>> parameters;

            public IEnumerable<KeyValuePair<string, string>> Parameters { get { return parameters; } }

            public IFacetManager Faceting { get; set; }

            public void AddParameter(string name, string value)
            {
                parameters.Add(new KeyValuePair<string, string>(name, value));
            }

            public SolrSearchQuery()
            {
                parameters = new List<KeyValuePair<string, string>>();
            }
        }


        /// <summary>
        /// Builds the query from the provided search criteria.
        /// </summary>
        /// <param name="criteria">The search criteria.</param>
        /// <returns>The query request object.</returns>
        public object BuildQuery(ISearchCriteria criteria)
        {
            SolrSearchQuery query = new SolrSearchQuery();

            query.AddParameter("q", BuildQueryString(criteria));

            SearchSortField[] sortFields = criteria.Sort == null ? null : criteria.Sort.GetSort();
            if (sortFields != null && sortFields.Length > 0)
            {
                StringBuilder sortBuilder = new StringBuilder();
                bool first = true;
                foreach (SearchSortField sort in sortFields)
                {
                    if (first) first = false;
                    else sortBuilder.Append(',');

                    sortBuilder.Append(sort.FieldName.ToLowerInvariant()).Append(sort.IsDescending ? " desc" : " asc");
                }

                query.AddParameter("sort", sortBuilder.ToString());
            }

            if (criteria.StartingRecord != 0) query.AddParameter("start", criteria.StartingRecord.ToString());
            query.AddParameter("rows", criteria.RecordsToRetrieve.ToString());

            query.Faceting = new FacetManager(criteria.Filters, criteria.Currency);

            return query;
        }


        private string BuildQueryString(ISearchCriteria criteria)
        {
            List<QueryPart> queryParts = new List<QueryPart>();

            queryParts.Add(new FieldQuery(BaseCatalogIndexBuilder.FieldConstants.Language, criteria.Locale.ToLower()));

            // call these property gets once only -- the ISearchCriteria implementation copies out a full new array for each access.
            string[] activeFilterFields = criteria.ActiveFilterFields;
            ISearchFilterValue[] activeFilterValues = criteria.ActiveFilterValues;
            if (activeFilterFields.Length != activeFilterValues.Length) throw new Exception("ActiveFilterFields and ActiveFilterValues have different lengths.");

            for (int i = 0; i < activeFilterFields.Length; ++i)
            {
                string fieldName = activeFilterFields[i].ToLowerInvariant();
                ISearchFilterValue value = activeFilterValues[i];
                SimpleValue simpleValue;
                RangeValue rangeValue;
                PriceRangeValue priceRangeValue;

                if ((simpleValue = (value as SimpleValue)) != null)
                {
                    queryParts.Add(new FieldQuery(fieldName, simpleValue.value));
                }
                else if ((rangeValue = (value as RangeValue)) != null)
                {
                    queryParts.Add(new RangeQuery(fieldName,
                        rangeValue.lowerbound, rangeValue.upperbound,
                        rangeValue.lowerboundincluded, rangeValue.upperboundincluded));
                }
                else if ((priceRangeValue = (value as PriceRangeValue)) != null)
                {
                    if (priceRangeValue.currency.Equals(criteria.Currency, StringComparison.OrdinalIgnoreCase))
                    {
                        queryParts.Add(new RangeQuery(fieldName + priceRangeValue.currency,
                            priceRangeValue.lowerbound, priceRangeValue.upperbound,
                            priceRangeValue.lowerboundincluded, priceRangeValue.upperboundincluded));
                    }
                }
            }

            SolrCatalogEntrySearchCriteria solrCriteria = criteria as SolrCatalogEntrySearchCriteria;
            CatalogEntrySearchCriteria catalogCriteria = solrCriteria != null ? solrCriteria.BaseCriteria : criteria as CatalogEntrySearchCriteria;
            if (catalogCriteria != null)
            {
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.Catalog, catalogCriteria.CatalogNames));
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.Node, catalogCriteria.CatalogNodes));
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.Outline, catalogCriteria.Outlines));
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.MetaClass, catalogCriteria.SearchIndex));
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.ClassType, catalogCriteria.ClassTypes.Cast<string>().Select(s => s.ToLowerInvariant())));
                
                if (solrCriteria == null || !solrCriteria.DisableDateFilter)
                {
                    // TODO: Change the date filter in ProductSearchProvider.cs when changing this. 
                    queryParts.Add(new RangeQuery(BaseCatalogIndexBuilder.FieldConstants.StartDate, MinDateTimeString, catalogCriteria.StartDate.ToString(SolrSearchProvider.DateTimeFormatString)));
                    queryParts.Add(new RangeQuery(BaseCatalogIndexBuilder.FieldConstants.EndDate, catalogCriteria.EndDate.ToString(SolrSearchProvider.DateTimeFormatString), MaxDateTimeString));
                }

                if (!string.IsNullOrEmpty(catalogCriteria.SearchPhrase))
                {
                    if (catalogCriteria.IsFuzzySearch)
                    {
                        // TODO: Verify this is correct functionality (issue 1939).
                        // TODO: When verified, move this out to an AddXxxQuery(state, ...) method.
                        StringBuilder searchPhraseBuilder = new StringBuilder();
                        foreach (string keyword in catalogCriteria.SearchPhrase.Split(' ').Where(s => !string.IsNullOrEmpty(s)))
                        {
                            searchPhraseBuilder.AppendFormat("{0}~{1}", keyword, catalogCriteria.FuzzyMinSimilarity);
                        }

                        if (searchPhraseBuilder.Length > 0) queryParts.Add(new VerbatimQuery(searchPhraseBuilder.ToString()));
                    }
                    else
                    {
                        queryParts.Add(new VerbatimQuery(catalogCriteria.SearchPhrase));
                    }
                }
            }

            // NOTE: Add any additional parts BEFORE the catalogCriteria.SearchPhrase handling.

            StringBuilder prefix = new StringBuilder();
            StringBuilder body = new StringBuilder();
            bool first = true;

            foreach (QueryPart part in queryParts.Where(p => p.HasContent()))
            {
                if (!first)
                {
                    prefix.Append('(');
                    body.Append(" AND ");
                }

                part.AppendTo(body);

                if (!first) body.Append(')');
                else first = false;
            }

            return prefix.Append(body).ToString();
        }
    }
}
